from django.conf import settings
from django.urls import include, path

from . import views

app_name = 'accounting'
urlpatterns = [
    path(r'book-incoming-invoice/', views.book_incoming_invoice, 'book_incoming_invoice'),
]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sepa', '0003_default_values'),
    ]

    operations = [
        migrations.AlterField(
            model_name='directdebitbatch',
            name='executed',
            field=models.DateTimeField(null=True, verbose_name='executed', blank=True),
        ),
        migrations.AlterField(
            model_name='directdebitmandate',
            name='reference',
            field=models.CharField(max_length=35, unique=True, null=True, verbose_name='mandate reference', blank=True),
        ),
    ]



# -*- coding: utf-8 -*-

from django.conf import settings
from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django_countries.fields import CountryField
from taggit.managers import TaggableManager

from shark.utils.fields import OldAddressField, AddressField, LanguageField
from shark.utils.id_generators import IdField
from shark.utils.settings import get_settings_instance, get_settings_value

NUMBER_GENERATOR = get_settings_instance('CUSTOMER.NUMBER_GENERATOR')
CUSTOMER_TYPE_CHOICES = get_settings_value('CUSTOMER.TYPE_CHOICES')
CUSTOMER_TYPE_DEFAULT = get_settings_value('CUSTOMER.TYPE_DEFAULT')


class CustomerTypeField(models.CharField):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_length', 20)
        kwargs.setdefault('choices', CUSTOMER_TYPE_CHOICES)
        kwargs.setdefault('default', CUSTOMER_TYPE_DEFAULT)
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["choices"]
        del kwargs["default"]
        return name, path, args, kwargs


class Customer(models.Model):
    number = IdField(generator=NUMBER_GENERATOR)
    # XXX add_unique constraint
    name = models.CharField(max_length=50, blank=True)
    # FIXME add choices
    type = CustomerTypeField(db_index=True)
    primary_admin = models.ForeignKey(settings.AUTH_USER_MODEL,
            on_delete=models.CASCADE,
            blank=True, null=True)

    address = OldAddressField(_('address'), blank=True) # XXX deprecated

    # Language to be used when communicating with the customer. This
    # field is mainly used to determine which language to use when
    # generating invoices and email messages.
    language = LanguageField(_('language'), blank=True)

    # rates when creating invoices
    hourly_rate = models.DecimalField(_('hourly rate'),
            max_digits=7, decimal_places=2, blank=True, null=True)
    daily_rate = models.DecimalField(_('daily rate'),
            max_digits=7, decimal_places=2, blank=True, null=True)

    tags = TaggableManager(blank=True)

    INVOICE_DISPATCH_TYPE_EMAIL = 'email'
    INVOICE_DISPATCH_TYPE_FAX = 'fax'
    INVOICE_DISPATCH_TYPE_MAIL = 'mail'
    # XXX move this to the billing app?
    INVOICE_DISPATCH_TYPE_CHOICES = (
        (INVOICE_DISPATCH_TYPE_EMAIL, _('via email')),
        (INVOICE_DISPATCH_TYPE_FAX, _('via FAX')),
        (INVOICE_DISPATCH_TYPE_MAIL, _('via mail')),
    )
    invoice_dispatch_type = models.CharField(max_length=20,
            choices=INVOICE_DISPATCH_TYPE_CHOICES,
            default='email',
            verbose_name=_('Invoice dispatch type'))
    PAYMENT_TYPE_INVOICE = 'invoice'
    PAYMENT_TYPE_DIRECT_DEBIT = 'direct_debit'
    PAYMENT_TYPE_CHOICES = (
        (PAYMENT_TYPE_INVOICE, _('Invoice')),
        (PAYMENT_TYPE_DIRECT_DEBIT, _('Direct debit')),
    )
    payment_type = models.CharField(max_length=20,
            choices=PAYMENT_TYPE_CHOICES,
            default='invoice',
            verbose_name=_('Payment Type'))
    vatin = models.CharField(max_length=14, blank=True,
            verbose_name=_('VATIN'),
            help_text=_('Value added tax identification number'))

    enabled = models.BooleanField(default=True)

    created = models.DateTimeField(_('created'), auto_now_add=True)
    updated = models.DateTimeField(_('updated'), auto_now=True)

    class Meta:
        verbose_name = _('customer')
        verbose_name_plural = _('customers')

    def __str__(self):
        return self.number

    @property
    def active(self):
        # The customer active flag does not depend on anything but
        # the enabled flag.
        return self.enabled

    @property
    def vat_required(self):
        # VAT for invoices is required if customer...
        # ...lives in Germany
        # ...lives in the EU and does not have a VATIN
        return self.country.id == 'DE' or \
                (self.country.eu and not self.vatin)


class CustomerComment(models.Model):
    customer = models.ForeignKey('customer.Customer',
            on_delete=models.CASCADE)
    user = models.ForeignKey(settings.AUTH_USER_MODEL,
            blank=True, null=True, on_delete=models.SET_NULL)
    text = models.TextField()

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)


class CustomerContact(models.Model):
    customer = models.ForeignKey('customer.Customer',
            on_delete=models.CASCADE)

    # TODO add type (person, general,... ?)
    GENDER_MALE = 'M'
    GENDER_FEMALE = 'F'
    GENDER_CHOICES = (
            (GENDER_MALE, _('Male')),
            (GENDER_FEMALE, _('Female')),
    )
    gender = models.CharField(max_length=1, blank=True,
            choices=GENDER_CHOICES,
            verbose_name=_('Gender'))
    TITLE_CHOICES = (
            ('Herr', 'Herr'),
            ('Frau', 'Frau'),
            ('Fräulein', 'Fräulein'),
            ('Dr.', 'Dr.'),
            ('Dr. Dr.', 'Dr. Dr.'),
            ('Prof. Dr.', 'Prof. Dr.'),
    )
    title = models.CharField(max_length=20, blank=True,
            choices = TITLE_CHOICES,
            verbose_name=_('Salutation'),
            help_text='z.B. Herr, Frau, Dr., Prof.,...')
    first_name = models.CharField(max_length=20, blank=True,
            verbose_name=_('First name'))
    last_name = models.CharField(max_length=20, blank=True,
            verbose_name=_('Last name'))

    email = models.EmailField(blank=True)
    phone_number = models.CharField(max_length=50, blank=True)
    mobile_number = models.CharField(max_length=50, blank=True)
    fax_number = models.CharField(max_length=50, blank=True)


class CustomerAddress(models.Model):
    customer = models.ForeignKey('customer.Customer', on_delete=models.CASCADE)
    address = AddressField(prefix='')
    sender_line = models.CharField(max_length=100, blank=True, default='')
    invoice_address = models.BooleanField(default=False)



# Credits:
# http://stackoverflow.com/questions/2267362/convert-integer-to-a-string-in-a-given-numeric-base-in-python

import string
digs = string.digits + string.ascii_lowercase

def int2base(x, base):
    if x < 0:
        sign = -1
    elif x == 0:
        return digs[0]
    else:
        sign = 1

    x *= sign
    digits = []

    while x:
        digits.append(digs[int(x % base)])
        x = int(x / base)

    if sign < 0:
        digits.append('-')

    digits.reverse()

    return ''.join(digits)



# Generated by Django 2.2.2 on 2019-06-05 15:21

from django.db import migrations, models
import django_countries.fields
import shark.utils.fields


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0003_meta_changes'),
    ]

    operations = [
        migrations.RenameField(
            model_name='invoice',
            old_name='recipient',
            new_name='old_recipient',
        ),
        migrations.AlterField(
            model_name='invoice',
            name='old_recipient',
            field=shark.utils.fields.OldAddressField(blank=True, db_column='recipient', help_text='This field will be automatically filled with the address of the customer.'),
        ),
        migrations.RenameField(
            model_name='invoice',
            old_name='sender',
            new_name='old_sender',
        ),
        migrations.AlterField(
            model_name='invoice',
            name='old_sender',
            field=shark.utils.fields.OldAddressField(blank=True, db_column='sender'),
        ),
        migrations.AddField(
            model_name='invoice',
            name='recipient_address_addition_1',
            field=models.CharField(blank=True, max_length=100, verbose_name='address addition (1st row)'),
        ),
        migrations.AddField(
            model_name='invoice',
            name='recipient_address_addition_2',
            field=models.CharField(blank=True, max_length=100, verbose_name='address addition (2nd row)'),
        ),
        migrations.AddField(
            model_name='invoice',
            name='recipient_city',
            field=models.CharField(blank=True, max_length=100, verbose_name='city'),
        ),
        migrations.AddField(
            model_name='invoice',
            name='recipient_country',
            field=django_countries.fields.CountryField(blank=True, max_length=2, verbose_name='country'),
        ),
        migrations.AddField(
            model_name='invoice',
            name='recipient_name',
            field=models.CharField(blank=True, max_length=100, verbose_name='name'),
        ),
        migrations.AddField(
            model_name='invoice',
            name='recipient_postal_code',
            field=models.CharField(blank=True, max_length=10, verbose_name='postal code'),
        ),
        migrations.AddField(
            model_name='invoice',
            name='recipient_state',
            field=models.CharField(blank=True, max_length=100),
        ),
        migrations.AddField(
            model_name='invoice',
            name='recipient_street',
            field=models.CharField(blank=True, max_length=100, verbose_name='street'),
        ),
        migrations.AddField(
            model_name='invoice',
            name='sender_address_addition_1',
            field=models.CharField(blank=True, max_length=100, verbose_name='address addition (1st row)'),
        ),
        migrations.AddField(
            model_name='invoice',
            name='sender_address_addition_2',
            field=models.CharField(blank=True, max_length=100, verbose_name='address addition (2nd row)'),
        ),
        migrations.AddField(
            model_name='invoice',
            name='sender_city',
            field=models.CharField(default='', max_length=100, verbose_name='city'),
        ),
        migrations.AddField(
            model_name='invoice',
            name='sender_country',
            field=django_countries.fields.CountryField(default='', max_length=2, verbose_name='country'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='invoice',
            name='sender_name',
            field=models.CharField(default='', max_length=100, verbose_name='name'),
        ),
        migrations.AddField(
            model_name='invoice',
            name='sender_postal_code',
            field=models.CharField(default='', max_length=10, verbose_name='postal code'),
        ),
        migrations.AddField(
            model_name='invoice',
            name='sender_state',
            field=models.CharField(blank=True, max_length=100),
        ),
        migrations.AddField(
            model_name='invoice',
            name='sender_street',
            field=models.CharField(default='', max_length=100, verbose_name='street'),
        ),
    ]



from os.path import basename, splitext
from tempfile import TemporaryFile

from django.core.files.storage import get_storage_class
from django.db import models
from django.db.models import signals
from django.dispatch import receiver
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django_hashedfilenamestorage.storage import HashedFilenameMetaStorage
import magic
from taggit.managers import TaggableManager
from wand.image import Image
from wand.exceptions import MissingDelegateError

from shark.utils.date import today
from shark.utils.fields import AddressField


DocumentStorage = HashedFilenameMetaStorage(storage_class=get_storage_class())


class Document(models.Model):
    title = models.CharField(_('title'),
            max_length=100)
    sender = AddressField(blank=True)
    recipient = AddressField(blank=True)
    vendor = models.ForeignKey('vendor.Vendor',
            verbose_name=_('vendor'),
            on_delete=models.SET_NULL,
            blank=True, null=True)
    customer = models.ForeignKey('customer.Customer',
            verbose_name=_('customer'),
            on_delete=models.SET_NULL,
            blank=True, null=True)
    TYPE_INVOICE = 'invoice'
    TYPE_PAYMENT_REMINDER = 'payment_reminder'
    TYPE_BANK_STATEMENT = 'bank_statement'
    TYPE_QUOTE = 'quote'
    TYPE_MISC = 'misc'
    TYPE_SEPA_DD_MANDATE = 'sepa_dd_mandate'
    TYPE_CHOICES = [
        (TYPE_INVOICE, _('invoice')),
        (TYPE_PAYMENT_REMINDER, _('payment reminder')),
        (TYPE_BANK_STATEMENT, _('bank statement')),
        (TYPE_QUOTE, _('quote')),
        (TYPE_MISC, _('miscellaneous')),
        (TYPE_SEPA_DD_MANDATE, _('SEPA direct debit mandate')),
    ]
    type = models.CharField(_('direction'), max_length=20, choices=TYPE_CHOICES)
    DIRECTION_INTERNAL = 'internal'
    DIRECTION_INBOUND = 'inbound'
    DIRECTION_OUTBOUND = 'outbound'
    DIRECTION_CHOICES = [
        (DIRECTION_INTERNAL, _('internal')),
        (DIRECTION_INBOUND, _('inbound')),
        (DIRECTION_OUTBOUND, _('outbound')),
    ]
    direction = models.CharField(_('direction'), max_length=10, choices=DIRECTION_CHOICES)
    date = models.DateField(_('date'), default=today,
            help_text='Date as written on the document.')
    file = models.FileField(_('file'),
            upload_to='documents', storage=DocumentStorage())
    original_filename = models.TextField(editable=False, blank=True, null=True)
    size = models.BigIntegerField(_('file size'),
            default=0, editable=False)
    mime_type = models.CharField(_('MIME type'),
            blank=True, max_length=100,
            help_text=_('Auto detected from uploaded file'))
    SOURCE_EMAIL = 'email'
    SOURCE_DOWNLOAD = 'download'
    SOURCE_MAIL = 'mail'
    SOURCE_FAX = 'fax'
    SOURCE_RECEIPT = 'receipt'
    SOURCE_SELF = 'self'
    SOURCE_CHOICES = [
        (SOURCE_EMAIL, _('email')),
        (SOURCE_DOWNLOAD, _('download')),
        (SOURCE_MAIL, _('mail')),
        (SOURCE_FAX, _('fax')),
        (SOURCE_RECEIPT, _('receipt')),
        (SOURCE_SELF, _('self')),
    ]
    source = models.CharField(_('original'),
            blank=True,
            max_length=10,
            choices=SOURCE_CHOICES,
            help_text=u'Where does this document come from?')
    comment = models.TextField(_('comment'),
            blank=True)
    tags = TaggableManager(_('tags'),
            blank=True)
    created = models.DateTimeField(_('created'), auto_now_add=True)
    received = models.DateField(_('received'), default=today,
            help_text='Date when the document was received.')

    THUMBNAIL_FIELDS = (
        ('thumbnail_small', 'sm', '128x128'),
        ('thumbnail_medium', 'md', '256x256'),
        ('thumbnail_large', 'lg', '512x512'),
    )
    thumbnail_small = models.ImageField(upload_to='documents',
            editable=False)
    thumbnail_medium = models.ImageField(upload_to='documents',
            editable=False)
    thumbnail_large = models.ImageField(upload_to='documents',
            editable=False)

    def __str__(self):
        return self.title

def create_thumbnails(doc):
    # create thumbnails
    fn, ext = splitext(basename(doc.file.name))
    fh = doc.file.open()
    try:
        with Image(file=fh) as orig:
            if len(orig.sequence) == 0:
                clear_thumbnails(doc)
            with Image(orig.sequence[0]) as thumb:
                thumb.format = 'png'
                for field_name, suffix, size in Document.THUMBNAIL_FIELDS:
                    thumb_clone = thumb.clone()
                    thumb_clone.transform('', size)
                    with TemporaryFile() as tmp:
                        thumb_clone.save(file=tmp)
                        tmp.seek(0)
                        field = getattr(doc, field_name)
                        field.save(f'{fn}_{suffix}.png', tmp, save=False)
    except MissingDelegateError:
        clear_thumbnails(doc)

def clear_thumbnails(doc):
    changed = False
    for field_name, suffix, size in Document.THUMBNAIL_FIELDS:
        field = getattr(doc, field_name)
        if field:
            getattr(doc, field_name).delete()
        changed = True
    return changed

@receiver(signals.pre_save, sender=Document)
def document_pre_save(instance, raw, **kwargs):
    if raw:
        return
    instance.size = instance.file.size
    instance.original_filename = instance.file.name
    instance.mime_type = magic.from_buffer(instance.file.read(1024), mime=True)
    instance._file_changed = not instance.pk or \
            Document.objects.get(pk=instance.pk).file != instance.file

@receiver(signals.post_save, sender=Document)
def document_post_save(instance, raw, **kwargs):
    if raw:
        return
    if instance._file_changed:
        create_thumbnails(instance)
        Document.objects.filter(pk=instance.pk).update(**{
            field_name: getattr(instance, field_name)
            for field_name, _unused, _unused in Document.THUMBNAIL_FIELDS
        })



from django.urls import path

from shark.sepa import admin_views

app_name = 'sepa'
urlpatterns = [
    path('directdebitbatch/<pk>/sepa.xml', admin_views.sepa_xml,
            name='directdebitbatch_sepaxml'),
]



# Generated by Django 2.2.2 on 2019-06-05 15:27

from django.db import migrations


country_map = {
    'SCHWEIZ': 'ch',
    'UNITED KINGDOM': 'gb',
    'DEUTSCHLAND': 'de'
}


def parse_address(s):
    lines = [l.strip() for l in s.split('\n')]
    # auto detect country
    if lines[-1].upper() in country_map:
        country = country_map[lines[-1].upper()]
        lines = lines[:-1]
    else:
        country = 'DE'
    try:
        postal_code, city = lines[-1].split(' ', 1)
    except ValueError:
        state = lines[-1]
        lines = lines[:-1]
    else:
        state = ''
    assert len(lines) <= 6
    return dict(
        name=lines[0],
        address_addition_1=lines[1] if len(lines) >= 4 else '',
        address_addition_2=lines[2] if len(lines) >= 5 else '',
        street=lines[-2],
        postal_code=lines[-1].split(' ', 1)[0],
        city=lines[-1].split(' ', 1)[1],
        state=state,
        country=country,
        invoice_address=True,
    )

def convert_data(apps, schema_editor):
    Invoice = apps.get_model('billing.Invoice')

    for invoice in Invoice.objects.all():
        for field_name, value in parse_address(invoice.old_recipient).items():
            setattr(invoice, f'recipient_{field_name}', value)
        for field_name, value in parse_address(invoice.old_sender).items():
            setattr(invoice, f'sender_{field_name}', value)
        invoice.save()


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0004_new_address_fields'),
    ]

    operations = [
        migrations.RunPython(convert_data),
    ]



from django.db import models


class Account(models.Model):
    name = models.CharField(max_lenght=50)


class Booking(models.Model):
    source = models.ForeignKey('accounting.Account', on_delete=models.CASCADE)
    target = models.ForeignKey('accounting.Account', on_delete=models.CASCADE)
    value = models.DecimalField(max_digits=12, decimal_places=2)
    text = models.CharField(max_length=200)
    # booking specific data (e.g. link to article, contract, etc.)
    reference = models.TextField()



from importlib import import_module


def dynamic_callable(name):
    module_name, function_name = name.rsplit('.', 1)
    def wrapped(*args, **kwargs):
        module = import_module(module_name)
        function = getattr(module, function_name)
        return function(*args, **kwargs)
    wrapped.__name__ == function_name
    return wrapped



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Account',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=50, verbose_name='account name')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Transaction',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('entry_date', models.DateField(verbose_name='entry date')),
                ('value_date', models.DateField(verbose_name='value date')),
                ('text_key', models.CharField(max_length=100)),
                ('primanota', models.CharField(max_length=100)),
                ('account_holder', models.CharField(max_length=100)),
                ('account_number', models.CharField(max_length=100)),
                ('bank_code', models.CharField(max_length=100)),
                ('reference', models.TextField(max_length=100)),
                ('currency', models.CharField(default=b'EUR', max_length=3)),
                ('amount', models.DecimalField(max_digits=11, decimal_places=2)),
                ('debit_credit', models.CharField(max_length=2)),
                ('account', models.ForeignKey(to='banking.Account', on_delete=models.CASCADE)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
    ]



from django.contrib import admin

from . import models


class RelationAInline(admin.TabularInline):
    model = models.Relation
    fk_name = 'a'
    extra = 0


class RelationBInline(admin.TabularInline):
    model = models.Relation
    fk_name = 'b'
    extra = 0


class ParticipantInline(admin.TabularInline):
    model = models.Participant
    extra = 0


class NoteInline(admin.StackedInline):
    model = models.Note
    extra = 0


@admin.register(models.Issue)
class IssueAdmin(admin.ModelAdmin):
    list_display = ('project', 'summary', 'created', 'updated')
    list_display_links = ('summary',)
    search_fields = ('project__name', 'summary')
    inlines = (
        RelationAInline,
        RelationBInline,
        ParticipantInline,
        NoteInline,
    )
    ordering = ('-updated',)


class AttachmentInline(admin.StackedInline):
    model = models.Attachment
    extra = 0


class NoteAdmin(admin.ModelAdmin):
    list_display = ('author', 'type', 'text', 'created')
    list_display_links = ('text',)
    search_fields = ('text',)
    ordering = ('-created',)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django_hashedfilenamestorage.storage
import taggit.managers
import shark.utils.date


class Migration(migrations.Migration):

    dependencies = [
        ('taggit', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Document',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('title', models.CharField(max_length=100, verbose_name='title')),
                ('date', models.DateField(default=shark.utils.date.today, help_text=b'Date as written on the document.', verbose_name='date')),
                ('file', models.FileField(upload_to=b'documents', storage=django_hashedfilenamestorage.storage.HashedFilenameFileSystemStorage(), verbose_name='file')),
                ('size', models.BigIntegerField(default=0, verbose_name='file size', editable=False)),
                ('mime_type', models.CharField(help_text='Auto detected from uploaded file', max_length=100, verbose_name='MIME type', blank=True)),
                ('original', models.CharField(help_text='Where does this document come from?', max_length=10, verbose_name='original', choices=[(b'email', 'email'), (b'download', 'download'), (b'mail', 'mail'), (b'fax', 'fax'), (b'receipt', 'receipt')])),
                ('comment', models.TextField(verbose_name='comment', blank=True)),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('received', models.DateField(default=shark.utils.date.today, help_text=b'Date when the document was received.', verbose_name='received')),
                ('tags', taggit.managers.TaggableManager(to='taggit.Tag', through='taggit.TaggedItem', blank=True, help_text='A comma-separated list of tags.', verbose_name='tags')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
    ]



# Generated by Django 2.2.2 on 2019-06-05 12:39

from django.db import migrations, models
import django.db.models.deletion
import django_countries.fields
import localflavor.generic.models
import shark.sepa.models


class Migration(migrations.Migration):

    dependencies = [
        ('sepa', '0005_change_due_date'),
    ]

    operations = [
        migrations.AlterField(
            model_name='directdebitbatch',
            name='creditor_iban',
            field=localflavor.generic.models.IBANField(default=shark.sepa.models.get_default_creditor_iban, include_countries=None, max_length=34, use_nordea_extensions=False, verbose_name='creditor IBAN'),
        ),
        migrations.AlterField(
            model_name='directdebitbatch',
            name='mandate_type',
            field=models.CharField(choices=[('CORE', 'CORE'), ('COR1', 'COR1'), ('B2B', 'B2B')], max_length=4, verbose_name='mandate type'),
        ),
        migrations.AlterField(
            model_name='directdebitbatch',
            name='sequence_type',
            field=models.CharField(choices=[('FRST', 'FRST'), ('RCUR', 'RCUR')], max_length=4, verbose_name='sequence type'),
        ),
        migrations.AlterField(
            model_name='directdebitmandate',
            name='bic',
            field=localflavor.generic.models.BICField(help_text='Bank Identifier Code', max_length=11, verbose_name='BIC'),
        ),
        migrations.AlterField(
            model_name='directdebitmandate',
            name='country',
            field=django_countries.fields.CountryField(default='DE', max_length=2),
        ),
        migrations.AlterField(
            model_name='directdebitmandate',
            name='iban',
            field=localflavor.generic.models.IBANField(help_text='International Bank Account Number', include_countries=None, max_length=34, use_nordea_extensions=False, verbose_name='IBAN'),
        ),
        migrations.AlterField(
            model_name='directdebitmandate',
            name='type',
            field=models.CharField(choices=[('CORE', 'CORE'), ('COR1', 'COR1'), ('B2B', 'B2B')], max_length=4),
        ),
        migrations.AlterField(
            model_name='directdebittransaction',
            name='mandate',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='sepa.DirectDebitMandate', verbose_name='SEPA DD mandate'),
        ),
    ]



# -*- coding: UTF-8 -*-

from datetime import date
from datetime import datetime
from decimal import Decimal

from django.conf import settings
from django.contrib import admin
from django.urls import reverse
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.utils.formats import date_format
from django.utils.html import format_html, format_html_join, mark_safe
from django.utils.translation import ugettext
from django.utils.translation import ugettext_lazy as _
from django.utils.translation import ungettext
import unicodecsv as csv

from shark import get_admin_change_url
from shark import get_admin_changelist_url
from shark.billing import models
from shark.customer.models import Customer


class excel_semicolon(csv.excel):
    delimiter = ';'


class InvoiceItemInline(admin.TabularInline):
    model = models.InvoiceItem
    extra = 3
    ordering = ('position',)
    exclude = ('customer',)


class InvoiceAdmin(admin.ModelAdmin):
    fieldsets = (
        (_('general'), {'fields': ('customer', 'type', 'number', 'language') }),
        (_('sender'), {'fields': [
            'sender_name',
            'sender_address_addition_1',
            'sender_address_addition_2',
            'sender_street',
            ('sender_postal_code', 'sender_city'),
            'sender_state',
            'sender_country',
        ]}),
        (_('recipient'), {'fields': [
            'recipient_name',
            'recipient_address_addition_1',
            'recipient_address_addition_2',
            'recipient_street',
            ('recipient_postal_code', 'recipient_city'),
            'recipient_state',
            'recipient_country',
        ]}),
        (_('dates'), {'fields': ('created', 'reminded', 'paid') }),
    )
    inlines = [InvoiceItemInline]
    list_display = ('number', 'get_customer', 'get_recipient', 'net',
            'gross', 'created', 'paid', 'is_okay', 'invoice_pdf', 'correction_pdf')
    list_editable = ('paid',)
    list_display_links = ('number',)
    list_select_related = True
    ordering = ('-created',)
    search_fields = ('number', 'customer__number', 'customer__name')
    list_filter = ('created', 'paid', 'type')
    date_hierarchy = 'created'
    actions = ('total_value_action', 'export_for_accounting')
    save_on_top = True
    autocomplete_fields = ('customer',)

    def get_customer(self, obj):
        return format_html('<a href="{}">{}</a>',
                get_admin_change_url(obj.customer),
                obj.customer)
    get_customer.short_description = _('Customer')
    get_customer.admin_order_field = 'customer'

    def get_recipient(self, obj):
        return format_html_join(mark_safe('<br>'), '{}', ((line,) for line in obj.recipient_lines))
    get_recipient.short_description = _('Recipient')

    def invoice_pdf(self, obj):
        view_url = reverse('billing_admin:invoice_pdf', args=(obj.number,))
        download_url = reverse('billing_admin:invoice_pdf', args=(obj.number,))
        return format_html('<a href="{}">View</a> | <a href="{}?download">Download</a>',
                view_url,
                download_url)
    invoice_pdf.short_description = 'Invoice'

    def correction_pdf(self, obj):
        view_url = reverse('billing_admin:correction_pdf', args=(obj.number,))
        download_url = reverse('billing_admin:correction_pdf', args=(obj.number,))
        return format_html('<a href="{}">View</a> | <a href="{}?download">Download</a>',
                view_url,
                download_url)
    correction_pdf.short_description = 'Correction'

    def response_add(self, request, obj, *args, **kwargs):
        obj.recalculate()
        obj.save()
        return super(InvoiceAdmin, self).response_add(request, obj, *args, **kwargs)

    def response_change(self, request, obj, *args, **kwargs):
        obj.recalculate()
        obj.save()
        return super(InvoiceAdmin, self).response_change(request, obj, *args, **kwargs)

    def total_value_action(self, request, queryset):
        net = Decimal(0)
        gross = Decimal(0)
        for invoice in queryset:
            net += invoice.net
            gross += invoice.gross
        value = ugettext('%(net)s net, %(gross)s gross') % {
                'net': net, 'gross': gross }
        self.message_user(request, ungettext(
            'Total value of %(count)d invoice: %(value)s',
            'Total value of %(count)d invoices: %(value)s',
            len(queryset)) % { 'count': len(queryset), 'value': value })
    total_value_action.short_description = _('Calculate total value of selected invoices')

    def export_for_accounting(self, request, queryset):
        queryset = queryset.select_related('customer')
        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = 'attachment; filename=invoices.csv'
        writer = csv.writer(response, dialect=excel_semicolon)
        cols = [
            ('created', None),
            ('paid', None),
            ('number', None),
            ('net', None),
            ('gross', None),
            ('vat_rate', lambda iv: iv.vat[0][0] if iv.vat else 0),
            ('vat', lambda iv: iv.vat[0][1] if iv.vat else 0),
            ('payment_type', None),
            ('customer_number', lambda iv: iv.customer.number),
            ('customer_vatin', lambda iv: iv.customer.vatin),
            ('address', lambda iv: iv.recipient), # FIXME replace by new customer.billing_address
        ]
        writer.writerow([c[0] for c in cols])
        for invoice in queryset:
            customer = invoice.customer
            writer.writerow([
                accessor(invoice) if accessor else getattr(invoice, name)
                for name, accessor in cols
            ])
        return response
    export_for_accounting.short_description = "Export selected invoices for accounting"



class InvoiceItemAdmin(admin.ModelAdmin):
    list_display = ('customer', 'invoice', 'position', 'sku', 'text',
            'begin', 'end', 'quantity', 'price', 'total', 'discount',
            'vat_rate')
    list_display_links = ('position', 'text')
    list_filter = ('begin', 'end', 'vat_rate')
    ordering = ('customer__number', 'invoice__number', 'position')
    search_fields = ('invoice__number', 'customer__number', 'sku', 'text')
    actions = ('action_create_invoice',)
    autocomplete_fields = ('customer', 'invoice')
    raw_id_fields = ('invoice',)

    def action_create_invoice(self, request, queryset):
        # customer_id -> items
        items_dict = {}
        # [(customer, items)]
        customer_items_list = []
        for item in queryset \
                .filter(invoice=None) \
                .order_by('text'):
            try:
                items = items_dict[item.customer_id]
            except KeyError:
                items = items_dict[item.customer_id] = []
                customer_items_list.append((item.customer, items))
            items.append(item)
        # create invoices
        for customer, items in customer_items_list:
            invoice = models.Invoice()
            invoice.customer = customer
            invoice.save()
            for position, item in enumerate(items, 1):
                item.position = position
                item.invoice = invoice
                item.save()
            invoice.recalculate()
            invoice.save()
        return HttpResponseRedirect(get_admin_changelist_url(models.Invoice))
    action_create_invoice.short_description = _('Create invoice(s) for selected item(s)')


admin.site.register(models.Invoice, InvoiceAdmin)
admin.site.register(models.InvoiceItem, InvoiceItemAdmin)



# -*- coding: UTF-8 -*-

from os.path import splitext

from django.contrib import admin
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _
from django.template.defaultfilters import filesizeformat

from shark.documents import models


class DocumentAdmin(admin.ModelAdmin):
    list_display = ('get_thumbnail', 'title', 'get_extension', 'date', 'get_size', 'mime_type')
    list_display_links = ('get_thumbnail', 'title')
    list_filter = ('date', 'source', 'mime_type', 'tags')
    search_fields = ('title',)

    def get_thumbnail(self, instance):
        if instance.thumbnail_small:
            return format_html('<img src="{}">', instance.thumbnail_small.url)
    get_thumbnail.short_description = _('Thumbnail')

    def get_extension(self, instance):
        fn, ext = splitext(instance.file.name)
        return ext[1:] if ext.startswith('.') else ext
    get_extension.short_description = _('Ext.')

    def get_size(self, instance):
        return filesizeformat(instance.size)
    get_size.short_description = _('Size')
    get_size.admin_order_field = 'size'


admin.site.register(models.Document, DocumentAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import taggit.managers
import shark.utils.fields
import django_countries.fields


class Migration(migrations.Migration):

    dependencies = [
        ('customer', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='customer',
            name='hourly_rate',
            field=models.DecimalField(null=True, verbose_name='hourly rate', max_digits=7, decimal_places=2, blank=True),
        ),
    ]



from django.urls import path

from . import admin_views as views


app_name = 'billing'
urlpatterns = [
# disabled for now
#    path('invoiceitem/invoice/', 'shark.billing.admin_views.invoice',
#            name='invoiceitem_invoice'),
    path('invoiceitem/import/', views.import_items,
            name='import_items'),
    path('invoice/<number>.pdf', views.invoice_pdf,
            name='invoice_pdf'),
    path('correction/<number>.pdf', views.correction_pdf,
            name='correction_pdf'),
]



# Generated by Django 2.2.2 on 2019-06-05 15:21

from django.db import migrations, models
import shark.utils.fields


class Migration(migrations.Migration):

    dependencies = [
        ('customer', '0006_add_address_addition_2'),
    ]

    operations = [
        migrations.AddField(
            model_name='customeraddress',
            name='state',
            field=models.CharField(blank=True, max_length=100),
        ),
        migrations.AlterField(
            model_name='customer',
            name='address',
            field=shark.utils.fields.OldAddressField(blank=True, verbose_name='address'),
        ),
        migrations.AlterField(
            model_name='customeraddress',
            name='address_addition_1',
            field=models.CharField(blank=True, max_length=100, verbose_name='address addition (1st row)'),
        ),
        migrations.AlterField(
            model_name='customeraddress',
            name='address_addition_2',
            field=models.CharField(blank=True, max_length=100, verbose_name='address addition (2nd row)'),
        ),
    ]



# -*- coding: UTF-8 -*-

from decimal import Decimal
from datetime import datetime, date, timedelta

from django.conf import settings
from django.db import models
from django.db.models import signals
from django.utils.formats import date_format
from django.utils.translation import ugettext_lazy as _

from shark.utils.settings import get_settings_value, get_settings_instance
from shark.utils.id_generators import IdField
from shark.utils.fields import OldAddressField, AddressField, LanguageField
from shark.utils.rounding import round_to_centi

INVOICE_PAYMENT_TIMEFRAME = get_settings_value('INVOICE.PAYMENT_TIMEFRAME', timedelta(days=14))
VAT_RATE_CHOICES = get_settings_value('VAT_RATE_CHOICES', ((Decimal(0), '0%'),))
INVOICE_SENDER = get_settings_value('INVOICE.SENDER')
UNIT_CHOICES = get_settings_value('INVOICE.UNIT_CHOICES')
NUMBER_GENERATOR = get_settings_instance('INVOICE.NUMBER_GENERATOR')


class Invoice(models.Model):
    #
    # general
    #
    customer = models.ForeignKey('customer.Customer', on_delete=models.CASCADE,
            verbose_name=_('Customer'))
    TYPE_INVOICE = 'invoice'
    TYPE_CORRECTION = 'correction'
    TYPE_CHOICES = (
        (TYPE_INVOICE, _('Invoice')),
        (TYPE_CORRECTION, _('Correction of invoice')),
    )
    type = models.CharField(_('type'),
            max_length=20,
            choices=TYPE_CHOICES,
            default=TYPE_INVOICE)
    number = IdField(verbose_name=_('number'),
            generator=NUMBER_GENERATOR)
    language = LanguageField(_('language'),
            blank=True)

    PAYMENT_TYPE_INVOICE = 'invoice'
    PAYMENT_TYPE_DIRECT_DEBIT = 'direct_debit'
    PAYMENT_TYPE_CHOICES = (
        (PAYMENT_TYPE_INVOICE, _('Invoice')),
        (PAYMENT_TYPE_DIRECT_DEBIT, _('Direct debit')),
    )
    payment_type = models.CharField(_('Payment Type'),
            max_length=20,
            choices=PAYMENT_TYPE_CHOICES,
            default='invoice',
            help_text=_('Will be copied from customer\'s preferences automatically.'))

    #
    # address
    #
    old_sender = OldAddressField(blank=True,
            db_column='sender')
    old_recipient = OldAddressField(blank=True,
            db_column='recipient',
            help_text=_('This field will be automatically filled with the address of the customer.'))

    sender = AddressField(default=INVOICE_SENDER)
    recipient = AddressField(blank=True)

    net = models.DecimalField(max_digits=10, decimal_places=2,
            default=Decimal('0.00'),
            verbose_name=_('net'))
    gross = models.DecimalField(max_digits=10, decimal_places=2,
            default=Decimal('0.00'),
            verbose_name=_('gross'))

    #
    # status
    #
    created = models.DateField(
            default=date.today,
            verbose_name=_('Created'))
    reminded = models.DateField(blank=True, null=True,
            verbose_name=_('Reminded'))
    paid = models.DateField(blank=True, null=True,
            verbose_name=_('Paid'))

    class Meta:
        db_table = 'billing_invoice'
        verbose_name = _('Invoice')
        verbose_name_plural = _('Invoices')
        unique_together = (('customer', 'number'),)
        ordering = ('-created',)

    def __str__(self):
        return '%s %s' % (_('Invoice'), self.number)

    def save(self, *args, **kwargs):
        if not self.recipient:
            self.recipient = self.customer.address
        if not self.language:
            self.language = self.customer.language \
                    if self.customer.language \
                    else settings.LANGUAGE_CODE
        super(Invoice, self).save(*args, **kwargs)

    @property
    def recipient_lines(self):
        return [
            line for line in [
                self.recipient.name,
                self.recipient.address_addition_1,
                self.recipient.address_addition_2,
                self.recipient.street,
                f'{self.recipient.postal_code} {self.recipient.city}',
                self.recipient.state,
                self.recipient.country.name if self.recipient.country != 'DE' else '',
            ] if line
        ]

    @property
    def sender_lines(self):
        return [
            line for line in [
                self.sender.name,
                self.sender.street,
                f'{self.sender.postal_code} {self.sender.city}',
                self.sender.country if self.sender.country != 'DE' else '',
            ] if line
        ]

    def is_okay(self):
        if self.paid:
            return True
        if self.reminded is None:
            deadline = self.created + INVOICE_PAYMENT_TIMEFRAME
        else:
            deadline = self.reminded + INVOICE_PAYMENT_TIMEFRAME
        return date.today() <= deadline
    is_okay.short_description = _('Okay')
    is_okay.boolean = True

    @property
    def items(self):
        if not hasattr(self, '_item_cache'):
            self._item_cache = self.item_set.all()
        return self._item_cache

    @property
    def correction(self):
        c = Invoice(
                customer=self.customer,
                number=self.number,
                language=self.language,
                sender=self.sender,
                recipient=self.recipient)
        c._item_cache = [item.clone() for item in self.items]
        for item in c.items:
            item.quantity = -item.quantity
        c.recalculate()
        return c

    #
    # status
    #
    created = models.DateField(
            default=date.today,
            verbose_name=_('Created'))
    reminded = models.DateField(blank=True, null=True,
            verbose_name=_('Reminded'))
    paid = models.DateField(blank=True, null=True,
            verbose_name=_('Paid'))

    def recalculate(self):
        self.net = sum(item.total for item in self.items)
        vat_amount = sum(vat_amount for vat_rate, vat_amount in self.vat)
        self.gross = self.net + vat_amount

    @property
    def vat(self):
        '''
        Return a list of (vat_rate, amount) tuples.
        '''
        # create a dict which maps the vat rate to a list of items
        vat_dict = {}
        for item in self.items:
            if item.vat_rate == 0:
                continue
            vat_dict.setdefault(item.vat_rate, []).append(item)
        # sum up item per vat rate and create an ordered list of
        # (vat_rate, vat_amount) tuples.
        vat_list = []
        for vat_rate, items in vat_dict.items():
            amount = sum(item.total for item in items)
            vat_amount = round_to_centi(vat_rate * amount)
            vat_list.append((vat_rate, vat_amount))
        vat_list.sort()
        return vat_list

    @property
    def vat_items(self):
        # XXX either this or the vat property should be dropped
        class VatItem(object):
            def __init__(self, rate, amount):
                self.rate = rate
                self.amount = amount
        return [VatItem(*t) for t in self.vat]


class UnitField(models.CharField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_length', 10)
        kwargs.setdefault('choices', UNIT_CHOICES)
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["choices"]
        if 'default' in kwargs:
            del kwargs["default"]
        return name, path, args, kwargs


class VatRateField(models.DecimalField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_digits', 3)
        kwargs.setdefault('decimal_places', 2)
        kwargs.setdefault('choices', VAT_RATE_CHOICES)
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["choices"]
        if 'default' in kwargs:
            del kwargs["default"]
        return name, path, args, kwargs



class InvoiceItem(models.Model):
    invoice = models.ForeignKey(Invoice, related_name='item_set', on_delete=models.CASCADE,
            blank=True, null=True,
            verbose_name=_('invoice'))
    customer = models.ForeignKey('customer.Customer', on_delete=models.CASCADE,
            verbose_name=_('customer'))
    TYPE_ITEM = 'item'
    TYPE_TITLE = 'title'
    TYPE_CHOICES = [
        (TYPE_ITEM, 'Item'),
        (TYPE_TITLE, 'Title'),
    ]
    type = models.CharField(max_length=10, choices=TYPE_CHOICES,
            default=TYPE_ITEM)
    position = models.PositiveIntegerField(blank=True, null=True,
            verbose_name=_('position'))
    quantity = models.DecimalField(max_digits=10, decimal_places=2,
            default=Decimal('1'),
            verbose_name=_('quantity'))
    sku = models.CharField(max_length=20, blank=True,
            verbose_name=_('SKU'),
            help_text=_('Stock-keeping unit (e.g. Article number)'))
    text = models.CharField(max_length=200,
            verbose_name=_('description'))
    begin = models.DateField(blank=True, null=True,
            verbose_name=_('begin'))
    end = models.DateField(blank=True, null=True,
            verbose_name=_('end'))
    price = models.DecimalField(max_digits=10, decimal_places=2,
            verbose_name=_('price'))
    unit = UnitField(blank=True, verbose_name=_('unit'))
    discount = models.DecimalField(max_digits=3, decimal_places=2,
            default=Decimal('0.00'),
            verbose_name=('discount'))
    vat_rate = VatRateField(
            verbose_name=('VAT rate'))

    class Meta:
        db_table = 'billing_invoice_item'
        verbose_name = _('Item')
        verbose_name_plural = _('Items')
        unique_together = (('invoice', 'position'),)
        ordering = ('position',)

    def __str__(self):
        return '#%d %s' % (self.position or 0, self.text)

    def clone(self):
        return InvoiceItem(
                invoice=self.invoice,
                customer=self.customer,
                position=self.position,
                quantity=self.quantity,
                sku=self.sku,
                text=self.text,
                begin=self.begin,
                end=self.end,
                price=self.price,
                unit=self.unit,
                discount=self.discount,
                vat_rate=self.vat_rate)

    def save(self):
        if not self.customer_id:
            if self.invoice_id:
                self.customer_id = self.invoice.customer_id
            else:
                raise RuntimeError('The customer must be set if no invoice is given')
        super(InvoiceItem, self).save()

    def get_period(self):
        if self.begin and self.end:
            begin = date_format(self.begin, 'SHORT_DATE_FORMAT')
            end = date_format(self.end, 'SHORT_DATE_FORMAT') \
                    if self.end is not None \
                    else _('one-time')
            return '%s – %s' % (begin, end)
        else:
            return None
    get_period.short_description = _('Billing period')
    period = property(get_period)

    @property
    def date(self):
        return self.begin or self.end \
                if bool(self.begin) != bool(self.end) \
                else None

    def get_subtotal(self):
        return round_to_centi(self.quantity * self.price)
    get_subtotal.short_description = _('Subtotal')
    subtotal = property(get_subtotal)

    @property
    def discount_percentage(self):
        return self.discount * 100

    @property
    def discount_amount(self):
        return round_to_centi(self.discount * self.subtotal)

    def get_total(self):
        return self.subtotal - self.discount_amount
    get_total.short_description = _('Sum of line')
    total = property(get_total)



# Generated by Django 2.2.2 on 2019-06-05 13:59

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('customer', '0004_add_various_fields'),
    ]

    operations = [
        migrations.AddField(
            model_name='customercontact',
            name='customer',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, to='customer.Customer'),
            preserve_default=False,
        ),
    ]



from django.conf import settings
from django.urls import reverse

__version__ = '0.0.1'


def get_admin_change_url(obj):
    meta = obj._meta
    app_label = meta.app_label
    model_name = meta.model_name
    view_name = 'admin:%s_%s_change' % (app_label, model_name)
    return reverse(view_name, args=(obj.pk,))


def get_admin_changelist_url(obj):
    meta = obj._meta
    app_label = meta.app_label
    model_name = meta.model_name
    view_name = 'admin:%s_%s_changelist' % (app_label, model_name)
    return reverse(view_name)



from django.utils import unittest

from shark.customer.models import Customer


class CustomerTestCase(unittest.TestCase):

    def test_get_address_lines(self):
        customer = Customer(address='foo\nbar')
        self.assertEquals(customer.get_address_lines(), ['foo', 'bar'])

    def test_set_address_lines(self):
        customer = Customer()
        customer.set_address_lines(['foo', 'bar'])
        self.assertEquals(customer.address, 'foo\nbar')

    def test_address_lines_property(self):
        customer = Customer(address_lines=['foo', 'bar'])
        self.assertEquals(customer.address_lines, ['foo', 'bar'])
        self.assertEquals(customer.address, 'foo\nbar')



import decimal

CENTI = decimal.Decimal('0.01')
DECIMAL_CONTEXT = decimal.DefaultContext.copy()
DECIMAL_CONTEXT.rounding = decimal.ROUND_HALF_UP


def round_to_centi(d):
    return DECIMAL_CONTEXT.quantize(d, CENTI)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('customer', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=100, verbose_name='name')),
                ('active', models.BooleanField(default=True, verbose_name='active')),
                ('customer', models.ForeignKey(verbose_name='customer', to='customer.Customer', on_delete=models.CASCADE)),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import datetime
from decimal import Decimal
import shark.utils.fields
import shark.utils.id_generators


class Migration(migrations.Migration):

    dependencies = [
        ('customer', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Invoice',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('type', models.CharField(default=b'invoice', max_length=20, choices=[(b'invoice', 'Invoice'), (b'correction', 'Correction of invoice')])),
                ('number', shark.utils.id_generators.IdField(unique=True, max_length=32, blank=True)),
                ('language', shark.utils.fields.LanguageField(blank=True, help_text='This field will be automatically filled with the language of the customer. If no language for the customer is specified the default language (en-us) will be used.', max_length=5, choices=[(b'af', b'Afrikaans'), (b'ar', b'Arabic'), (b'ast', b'Asturian'), (b'az', b'Azerbaijani'), (b'bg', b'Bulgarian'), (b'be', b'Belarusian'), (b'bn', b'Bengali'), (b'br', b'Breton'), (b'bs', b'Bosnian'), (b'ca', b'Catalan'), (b'cs', b'Czech'), (b'cy', b'Welsh'), (b'da', b'Danish'), (b'de', b'German'), (b'el', b'Greek'), (b'en', b'English'), (b'en-au', b'Australian English'), (b'en-gb', b'British English'), (b'eo', b'Esperanto'), (b'es', b'Spanish'), (b'es-ar', b'Argentinian Spanish'), (b'es-mx', b'Mexican Spanish'), (b'es-ni', b'Nicaraguan Spanish'), (b'es-ve', b'Venezuelan Spanish'), (b'et', b'Estonian'), (b'eu', b'Basque'), (b'fa', b'Persian'), (b'fi', b'Finnish'), (b'fr', b'French'), (b'fy', b'Frisian'), (b'ga', b'Irish'), (b'gl', b'Galician'), (b'he', b'Hebrew'), (b'hi', b'Hindi'), (b'hr', b'Croatian'), (b'hu', b'Hungarian'), (b'ia', b'Interlingua'), (b'id', b'Indonesian'), (b'io', b'Ido'), (b'is', b'Icelandic'), (b'it', b'Italian'), (b'ja', b'Japanese'), (b'ka', b'Georgian'), (b'kk', b'Kazakh'), (b'km', b'Khmer'), (b'kn', b'Kannada'), (b'ko', b'Korean'), (b'lb', b'Luxembourgish'), (b'lt', b'Lithuanian'), (b'lv', b'Latvian'), (b'mk', b'Macedonian'), (b'ml', b'Malayalam'), (b'mn', b'Mongolian'), (b'mr', b'Marathi'), (b'my', b'Burmese'), (b'nb', b'Norwegian Bokmal'), (b'ne', b'Nepali'), (b'nl', b'Dutch'), (b'nn', b'Norwegian Nynorsk'), (b'os', b'Ossetic'), (b'pa', b'Punjabi'), (b'pl', b'Polish'), (b'pt', b'Portuguese'), (b'pt-br', b'Brazilian Portuguese'), (b'ro', b'Romanian'), (b'ru', b'Russian'), (b'sk', b'Slovak'), (b'sl', b'Slovenian'), (b'sq', b'Albanian'), (b'sr', b'Serbian'), (b'sr-latn', b'Serbian Latin'), (b'sv', b'Swedish'), (b'sw', b'Swahili'), (b'ta', b'Tamil'), (b'te', b'Telugu'), (b'th', b'Thai'), (b'tr', b'Turkish'), (b'tt', b'Tatar'), (b'udm', b'Udmurt'), (b'uk', b'Ukrainian'), (b'ur', b'Urdu'), (b'vi', b'Vietnamese'), (b'zh-cn', b'Simplified Chinese'), (b'zh-hans', b'Simplified Chinese'), (b'zh-hant', b'Traditional Chinese'), (b'zh-tw', b'Traditional Chinese')])),
                ('sender', shark.utils.fields.OldAddressField(default=b"settings.SHARK['INVOICE']['SENDER']", blank=True)),
                ('recipient', shark.utils.fields.OldAddressField(help_text='This field will be automatically filled with the address of the customer.', blank=True)),
                ('net', models.DecimalField(default=Decimal('0.00'), verbose_name='net', max_digits=10, decimal_places=2)),
                ('gross', models.DecimalField(default=Decimal('0.00'), verbose_name='gross', max_digits=10, decimal_places=2)),
                ('created', models.DateField(default=datetime.date.today, verbose_name='Created')),
                ('reminded', models.DateField(null=True, verbose_name='Reminded', blank=True)),
                ('paid', models.DateField(null=True, verbose_name='Paid', blank=True)),
                ('customer', models.ForeignKey(verbose_name='Customer', to='customer.Customer', on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ('-created',),
                'db_table': 'billing_invoice',
                'verbose_name': 'Invoice',
                'verbose_name_plural': 'Invoices',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='InvoiceItem',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('position', models.PositiveIntegerField(null=True, verbose_name='position', blank=True)),
                ('quantity', models.DecimalField(default=Decimal('1'), verbose_name='quantity', max_digits=10, decimal_places=2)),
                ('sku', models.CharField(help_text='Stock-keeping unit (e.g. Article number)', max_length=20, verbose_name='SKU', blank=True)),
                ('text', models.CharField(max_length=200, verbose_name='description')),
                ('begin', models.DateField(null=True, verbose_name='begin', blank=True)),
                ('end', models.DateField(null=True, verbose_name='end', blank=True)),
                ('price', models.DecimalField(verbose_name='price', max_digits=10, decimal_places=2)),
                ('unit', models.CharField(blank=True, max_length=10, verbose_name='unit', choices=[(b's', 'second [s]'), (b'min', 'minute [min]'), (b'h', 'hour [h]'), (b'd', 'day [d]'), (b'w', 'week [w]'), (b'm', 'month [m]'), (b'a', 'year [a]')])),
                ('discount', models.DecimalField(default=Decimal('0.00'), verbose_name=b'discount', max_digits=3, decimal_places=2)),
                ('vat_rate', models.DecimalField(verbose_name=b'VAT rate', max_digits=3, decimal_places=2, choices=[(Decimal('0.19'), '19%'), (Decimal('0.07'), '7%'), (Decimal('0.00'), '0%% (tax free)')])),
                ('customer', models.ForeignKey(verbose_name='customer', to='customer.Customer', on_delete=models.CASCADE)),
                ('invoice', models.ForeignKey(related_name='item_set', verbose_name='invoice', blank=True, to='billing.Invoice', null=True, on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ('position',),
                'db_table': 'billing_invoice_item',
                'verbose_name': 'Item',
                'verbose_name_plural': 'Items',
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='invoiceitem',
            unique_together=set([('invoice', 'position')]),
        ),
        migrations.AlterUniqueTogether(
            name='invoice',
            unique_together=set([('customer', 'number')]),
        ),
    ]



import inspect

from django.conf import settings

from shark.utils.importlib import import_object

RAISE_ERROR = object()


def get_settings_value(name, default=RAISE_ERROR):
    d = settings.SHARK
    parts = name.split('.')
    for part in parts:
        try:
            d = d[part]
        except KeyError:
            if default is RAISE_ERROR:
                raise KeyError('No such setting: settings.SHARK%s' % ''.join('[%r]' % part for part in parts))
            return default
    return d


def shark_settings(d, base=None):
    if not base:
        from shark import settings_base
        base = settings_base.SHARK
    settings = base.copy()
    for key, value in d.items():
        if isinstance(value, dict):
            if not isinstance(base.get(key, {}), dict):
                raise RuntimeError('Type mismatch of settings and base settings: %r' % key)
            settings[key] = shark_settings(value, base.get(key, {}))
        else:
            settings[key] = value
    return settings


def get_settings_instance(name):
    v = get_settings_value(name)
    if isinstance(v, str):
        v = import_object(v)
    if inspect.isclass(v):
        v = v()
    return v



from django.urls import path
from rest_framework.urlpatterns import format_suffix_patterns

from . import api_views as views


app_name ='billing'
urlpatterns = [
    path(r'invoice/', views.InvoiceList.as_view(), name='invoice_list'),
    path(r'invoice/create/', views.InvoiceCreate.as_view(), name='invoice_create'),
    path(r'invoice/<int:pk>/', views.InvoiceDetail.as_view(), name='invoice_detail'),
]

urlpatterns = format_suffix_patterns(urlpatterns)



# Generated by Django 2.2.2 on 2019-06-05 15:21

from django.db import migrations, models
import django_countries.fields


class Migration(migrations.Migration):

    dependencies = [
        ('documents', '0004_improve_document'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='document',
            name='recipient',
        ),
        migrations.RemoveField(
            model_name='document',
            name='sender',
        ),
        migrations.AddField(
            model_name='document',
            name='recipient_address_addition_1',
            field=models.CharField(blank=True, max_length=100, verbose_name='address addition (1st row)'),
        ),
        migrations.AddField(
            model_name='document',
            name='recipient_address_addition_2',
            field=models.CharField(blank=True, max_length=100, verbose_name='address addition (2nd row)'),
        ),
        migrations.AddField(
            model_name='document',
            name='recipient_city',
            field=models.CharField(blank=True, max_length=100, verbose_name='city'),
        ),
        migrations.AddField(
            model_name='document',
            name='recipient_country',
            field=django_countries.fields.CountryField(blank=True, max_length=2, verbose_name='country'),
        ),
        migrations.AddField(
            model_name='document',
            name='recipient_name',
            field=models.CharField(blank=True, max_length=100, verbose_name='name'),
        ),
        migrations.AddField(
            model_name='document',
            name='recipient_postal_code',
            field=models.CharField(blank=True, max_length=10, verbose_name='postal code'),
        ),
        migrations.AddField(
            model_name='document',
            name='recipient_state',
            field=models.CharField(blank=True, max_length=100),
        ),
        migrations.AddField(
            model_name='document',
            name='recipient_street',
            field=models.CharField(blank=True, max_length=100, verbose_name='street'),
        ),
        migrations.AddField(
            model_name='document',
            name='sender_address_addition_1',
            field=models.CharField(blank=True, max_length=100, verbose_name='address addition (1st row)'),
        ),
        migrations.AddField(
            model_name='document',
            name='sender_address_addition_2',
            field=models.CharField(blank=True, max_length=100, verbose_name='address addition (2nd row)'),
        ),
        migrations.AddField(
            model_name='document',
            name='sender_city',
            field=models.CharField(blank=True, max_length=100, verbose_name='city'),
        ),
        migrations.AddField(
            model_name='document',
            name='sender_country',
            field=django_countries.fields.CountryField(blank=True, max_length=2, verbose_name='country'),
        ),
        migrations.AddField(
            model_name='document',
            name='sender_name',
            field=models.CharField(blank=True, max_length=100, verbose_name='name'),
        ),
        migrations.AddField(
            model_name='document',
            name='sender_postal_code',
            field=models.CharField(blank=True, max_length=10, verbose_name='postal code'),
        ),
        migrations.AddField(
            model_name='document',
            name='sender_state',
            field=models.CharField(blank=True, max_length=100),
        ),
        migrations.AddField(
            model_name='document',
            name='sender_street',
            field=models.CharField(blank=True, max_length=100, verbose_name='street'),
        ),
    ]



# Generated by Django 2.2.2 on 2019-06-05 12:51

from decimal import Decimal
from django.db import migrations, models
import shark.billing.models
import shark.utils.fields


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0002_add_item_type_and_payment_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='invoice',
            name='payment_type',
            field=models.CharField(choices=[('invoice', 'Invoice'), ('direct_debit', 'Direct debit')], default='invoice', help_text="Will be copied from customer's preferences automatically.", max_length=20, verbose_name='Payment Type'),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='sender',
            field=shark.utils.fields.OldAddressField(blank=True, default='Awesome Corp\nGreatstreet 123\n12345 Nicertown'),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='type',
            field=models.CharField(choices=[('invoice', 'Invoice'), ('correction', 'Correction of invoice')], default='invoice', max_length=20, verbose_name='type'),
        ),
        migrations.AlterField(
            model_name='invoiceitem',
            name='discount',
            field=models.DecimalField(decimal_places=2, default=Decimal('0.00'), max_digits=3, verbose_name='discount'),
        ),
        migrations.AlterField(
            model_name='invoiceitem',
            name='type',
            field=models.CharField(choices=[('item', 'Item'), ('title', 'Title')], default='item', max_length=10),
        ),
        migrations.AlterField(
            model_name='invoiceitem',
            name='unit',
            field=shark.billing.models.UnitField(blank=True, max_length=10, verbose_name='unit'),
        ),
        migrations.AlterField(
            model_name='invoiceitem',
            name='vat_rate',
            field=shark.billing.models.VatRateField(decimal_places=2, max_digits=3, verbose_name='VAT rate'),
        ),
    ]



# -*- coding: utf-8 -*-

from decimal import Decimal

from django.urls import reverse
from rest_framework import status
from rest_framework.test import APITestCase

from .models import Invoice


class AccountTests(APITestCase):

    def test_create_invoice(self):
        """
        Ensure we can create a new account object.
        """
        url = reverse('api:billing:invoice_create')
        response = self.client.post(url, {
            'customer': {
                #'number': test_customer.number,
                'number': 'TEST',
            },
            'number': 'TEST-0001',
            'items': [
                {
                    "position": 1,
                    "quantity": "100.00",
                    "sku": "",
                    "text": "Waffelhörnchen",
                    "begin": None,
                    "end": "2015-01-01",
                    "price": "0.1",
                    "unit": "m",
                    "discount": "0.00",
                    "vat_rate": 0.19
                }
            ]
        }, format='json')
        print response
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        # check the newly created invoice
        self.assertEqual(Invoice.objects.count(), 1)
        invoice = Invoice.objects.count
        customer = invoice.customer
        self.assertEqual(customer.number, 'TEST')
        self.assertEqual(invoice.number, 'TEST-0001')
        self.assertEqual(Invoice.objects.get().net, Decimal('10.00'))
        self.assertEqual(Invoice.objects.get().gross, Decimal('11.90'))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import localflavor.generic.models
import shark.sepa.models


class Migration(migrations.Migration):

    dependencies = [
        ('sepa', '0002_directdebitmandate_bank_name'),
    ]

    operations = [
        migrations.AlterField(
            model_name='directdebitbatch',
            name='creditor_bic',
            field=localflavor.generic.models.BICField(default=shark.sepa.models.get_default_creditor_bic, max_length=11, verbose_name='creditor BIC'),
        ),
        migrations.AlterField(
            model_name='directdebitbatch',
            name='creditor_country',
            field=models.CharField(default=shark.sepa.models.get_default_creditor_country, max_length=2, verbose_name='creditor country'),
        ),
        migrations.AlterField(
            model_name='directdebitbatch',
            name='creditor_iban',
            field=localflavor.generic.models.IBANField(default=shark.sepa.models.get_default_creditor_iban, max_length=34),
        ),
        migrations.AlterField(
            model_name='directdebitbatch',
            name='creditor_id',
            field=models.CharField(default=shark.sepa.models.get_default_creditor_id, max_length=20, verbose_name='creditor id'),
        ),
        migrations.AlterField(
            model_name='directdebitbatch',
            name='creditor_name',
            field=models.CharField(default=shark.sepa.models.get_default_creditor_name, max_length=70, verbose_name='creditor name'),
        ),
    ]



# -*- encoding: utf-8 -*-

DATE_FORMAT = 'j. F Y'
TIME_FORMAT = 'H:i:s'
DATETIME_FORMAT = 'j. F Y, H:i:s'
YEAR_MONTH_FORMAT = 'F Y'
MONTH_DAY_FORMAT = 'j. F'
SHORT_DATE_FORMAT = 'Y-m-d'
SHORT_DATETIME_FORMAT = 'Y-m-d H:i:s'
FIRST_DAY_OF_WEEK = 1 # Sunday
DATE_INPUT_FORMATS = (
    '%Y-%m-%d', '%y-%m-%d',     # '2006-10-25', '06-10-25'
)
TIME_INPUT_FORMATS = (
    '%H:%M:%S',     # '14:30:59'
    '%H:%M',        # '14:30'
)
DATETIME_INPUT_FORMATS = (
    '%Y-%m-%d %H:%M:%S',     # '2006-10-25 14:30:59'
    '%Y-%m-%d %H:%M',        # '2006-10-25 14:30'
    '%Y-%m-%d',              # '2006-10-25'
)
DECIMAL_SEPARATOR = '.'
THOUSAND_SEPARATOR = ','
NUMBER_GROUPING = 3



from django.contrib.auth.decorators import permission_required
from django.http import HttpResponse
from django.shortcuts import get_object_or_404

from . import models


@permission_required('sepa.change_directdebitbatch')
def sepa_xml(self, pk):
    dd_batch = get_object_or_404(models.DirectDebitBatch, pk=pk)
    response = HttpResponse(dd_batch.render_sepa_xml(), content_type='application/xml')
    response['Content-Disposition'] = 'attachment; filename=SEPA_DD_%s.xml' % dd_batch.uuid
    return response



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0006_require_contenttypes_0002'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('project', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Attachment',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('file', models.FileField(upload_to=b'issue/attachment', verbose_name='file')),
                ('description', models.TextField(verbose_name='description', blank=True)),
            ],
        ),
        migrations.CreateModel(
            name='Issue',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('summary', models.CharField(help_text='A concise description of the problem limited to 200 characters.', max_length=200, verbose_name='summary')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='updated')),
                ('project', models.ForeignKey(verbose_name='project', to='project.Project', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name': 'issue',
                'verbose_name_plural': 'issues',
            },
        ),
        migrations.CreateModel(
            name='Note',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('type', models.CharField(max_length=10, verbose_name='type')),
                ('text', models.TextField()),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('author', models.ForeignKey(verbose_name='author', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
                ('issue', models.ForeignKey(verbose_name='note', to='issue.Issue', on_delete=models.CASCADE)),
            ],
        ),
        migrations.CreateModel(
            name='Participant',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('watcher', models.BooleanField(default=False)),
                ('reporter', models.BooleanField(default=False)),
                ('waiting', models.BooleanField(default=False)),
                ('group', models.ForeignKey(verbose_name='group', to='auth.Group', on_delete=models.CASCADE)),
                ('issue', models.ForeignKey(verbose_name='issue', to='issue.Issue', on_delete=models.CASCADE)),
                ('user', models.ForeignKey(verbose_name='user', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name': 'participant',
                'verbose_name_plural': 'participants',
            },
        ),
        migrations.CreateModel(
            name='Relation',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('b_is_parent', models.BooleanField()),
                ('b_is_related', models.BooleanField()),
                ('b_is_duplicate', models.BooleanField()),
                ('type', models.CharField(max_length=10, verbose_name='type', choices=[(b'parent', 'parent'), (b'related', 'related'), (b'duplicate', 'duplicate')])),
                ('a', models.ForeignKey(related_name='related_b_set', to='issue.Issue', on_delete=models.CASCADE)),
                ('b', models.ForeignKey(related_name='related_a_set', to='issue.Issue', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name': 'relation',
                'verbose_name_plural': 'relations',
            },
        ),
        migrations.AddField(
            model_name='attachment',
            name='note',
            field=models.ForeignKey(verbose_name='note', to='issue.Note', on_delete=models.CASCADE),
        ),
        migrations.AlterUniqueTogether(
            name='relation',
            unique_together=set([('a', 'b')]),
        ),
        migrations.AlterUniqueTogether(
            name='participant',
            unique_together=set([('issue', 'user'), ('issue', 'group')]),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import uuid

from django.db import models, migrations
import django_countries.fields
import localflavor.generic.models


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0001_initial'),
        ('documents', '0001_initial'),
        ('customer', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='DirectDebitBatch',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('uuid', models.UUIDField(default=uuid.uuid4, verbose_name='UUID')),
                ('creditor_id', models.CharField(default=b'', max_length=20, verbose_name='creditor id')),
                ('creditor_name', models.CharField(default=b'', max_length=70, verbose_name='creditor name')),
                ('creditor_country', models.CharField(default=b'DE', max_length=2, verbose_name='creditor country')),
                ('creditor_iban', localflavor.generic.models.IBANField(default=b'', max_length=34)),
                ('creditor_bic', localflavor.generic.models.BICField(default=b'', max_length=11, verbose_name='creditor BIC')),
                ('due_date', models.DateTimeField(help_text='Must be min. 5 TARGET dates in the future for the first transaction and 2 target days in the future for recurring transactions.', verbose_name='due date')),
                ('mandate_type', models.CharField(max_length=4, verbose_name='mandate type', choices=[(b'CORE', b'CORE'), (b'COR1', b'COR1'), (b'B2B', b'B2B')])),
                ('sequence_type', models.CharField(max_length=4, verbose_name='sequence type', choices=[(b'FRST', b'FRST'), (b'RCUR', b'RCUR')])),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('executed', models.DateTimeField(verbose_name='executed')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='DirectDebitMandate',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('reference', models.CharField(unique=True, max_length=35, verbose_name='mandate reference', blank=True)),
                ('name', models.CharField(max_length=100)),
                ('street', models.CharField(max_length=100)),
                ('postal_code', models.CharField(max_length=20)),
                ('city', models.CharField(max_length=100)),
                ('country', django_countries.fields.CountryField(default=b'DE', max_length=2)),
                ('iban', localflavor.generic.models.IBANField(help_text=b'International Bank Account Number', max_length=34)),
                ('bic', localflavor.generic.models.BICField(help_text=b'Bank Identifier Code', max_length=11, verbose_name=b'BIC')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='updated')),
                ('signed', models.DateField(null=True, blank=True)),
                ('revoked', models.DateField(null=True, verbose_name='revoked', blank=True)),
                ('last_used', models.DateField(null=True, verbose_name='last_used', blank=True)),
                ('type', models.CharField(max_length=4, choices=[(b'CORE', b'CORE'), (b'COR1', b'COR1'), (b'B2B', b'B2B')])),
                ('customer', models.ForeignKey(related_name='direct_debit_mandate_set', verbose_name='customer', to='customer.Customer', on_delete=models.CASCADE)),
                ('document', models.ForeignKey(verbose_name='signed document', blank=True, to='documents.Document', null=True, on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name': 'SEPA direct debit mandate',
                'verbose_name_plural': 'SEPA direct debit mandates',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='DirectDebitTransaction',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('reference', models.CharField(max_length=140)),
                ('amount', models.DecimalField(verbose_name='amount', max_digits=11, decimal_places=2)),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('batch', models.ForeignKey(verbose_name='SEPA DD batch', to='sepa.DirectDebitBatch', on_delete=models.CASCADE)),
                ('customer', models.ForeignKey(related_name='direct_debit_transaction_set', verbose_name='customer', to='customer.Customer', on_delete=models.CASCADE)),
                ('invoice', models.ForeignKey(verbose_name='invoice', blank=True, to='billing.Invoice', null=True, on_delete=models.CASCADE)),
                ('mandate', models.ForeignKey(verbose_name=b'SEPA DD mandate', to='sepa.DirectDebitMandate', on_delete=models.CASCADE)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
    ]



from django.db import models
from django.utils.translation import ugettext_lazy as _


class Issue(models.Model):
    project = models.ForeignKey('project.Project', verbose_name=_('project'), on_delete=models.CASCADE)
    summary = models.CharField(_('summary'), max_length=200,
            help_text=_('A concise description of the problem limited to 200 characters.'))
    created = models.DateTimeField(_('created'), auto_now_add=True)
    updated = models.DateTimeField(_('updated'), auto_now=True)

    class Meta:
        verbose_name = _('issue')
        verbose_name_plural = _('issues')

    def __str__(self):
        return '[%s] %s' % (self.project, self.summary)


class Relation(models.Model):
    a = models.ForeignKey('issue.Issue', related_name='related_b_set', on_delete=models.CASCADE)
    b = models.ForeignKey('issue.Issue', related_name='related_a_set', on_delete=models.CASCADE)
    b_is_parent = models.BooleanField()
    b_is_related = models.BooleanField()
    b_is_duplicate = models.BooleanField()
    TYPE_PARENT = 'parent'
    TYPE_RELATED = 'related'
    TYPE_DUPLICATE = 'duplicate'
    TYPE_CHOICES = (
        # a has a parent b
        (TYPE_PARENT, _('parent')),
        # a has a related issue b
        (TYPE_RELATED, _('related')),
        # a has a duplicate b
        (TYPE_DUPLICATE, _('duplicate')),
    )
    type = models.CharField(_('type'), max_length=10, choices=TYPE_CHOICES)

    class Meta:
        verbose_name = _('relation')
        verbose_name_plural = _('relations')
        unique_together = (('a', 'b'),)


class Participant(models.Model):
    issue = models.ForeignKey('issue.Issue', verbose_name=_('issue'), on_delete=models.CASCADE)
    user = models.ForeignKey('auth.User', verbose_name=_('user'), on_delete=models.CASCADE)
    group = models.ForeignKey('auth.Group', verbose_name=_('group'), on_delete=models.CASCADE)
    watcher = models.BooleanField(default=False)
    reporter = models.BooleanField(default=False)
    waiting = models.BooleanField(default=False)

    class Meta:
        verbose_name = _('participant')
        verbose_name_plural = _('participants')
        unique_together = (('issue', 'user'), ('issue', 'group'))


class Note(models.Model):
    issue = models.ForeignKey('issue.Issue', verbose_name=_('note'), on_delete=models.CASCADE)
    author = models.ForeignKey('auth.User', verbose_name=_('author'), on_delete=models.CASCADE)
    TYPE_QUESTION = 'question'
    TYPE_BUG = 'question'
    TYPE_TODO = 'todo'
    TYPE_FEATURE = 'feature'
    TYPE_CHOICES = (
        (TYPE_QUESTION, _('question')),
        (TYPE_BUG, _('bug')),
        (TYPE_FEATURE, _('feature')),
        (TYPE_TODO, _('feature')),
    )
    type = models.CharField(_('type'), max_length=10)
    text = models.TextField()
    created = models.DateTimeField(_('created'), auto_now_add=True)


class Attachment(models.Model):
    note = models.ForeignKey('issue.Note', verbose_name=_('note'), on_delete=models.CASCADE)
    file = models.FileField(_('file'), upload_to='issue/attachment')
    description = models.TextField(_('description'), blank=True)



# Generated by Django 2.2.2 on 2019-06-05 12:39

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import shark.customer.models
import shark.utils.fields
import taggit.managers


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('customer', '0003_extra_fields_and_customer_address'),
    ]

    operations = [
        migrations.AddField(
            model_name='customer',
            name='name',
            field=models.CharField(blank=True, max_length=50),
        ),
        migrations.AddField(
            model_name='customer',
            name='primary_admin',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='customer',
            name='type',
            field=shark.customer.models.CustomerTypeField(db_index=True, max_length=20),
        ),
        migrations.AddField(
            model_name='customeraddress',
            name='sender_line',
            field=models.CharField(blank=True, default='', max_length=100),
        ),
        migrations.AddField(
            model_name='customercomment',
            name='user',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='customercontact',
            name='email',
            field=models.EmailField(blank=True, max_length=254),
        ),
        migrations.AlterField(
            model_name='customer',
            name='invoice_dispatch_type',
            field=models.CharField(choices=[('email', 'via email'), ('fax', 'via FAX'), ('mail', 'via mail')], default='email', max_length=20, verbose_name='Invoice dispatch type'),
        ),
        migrations.AlterField(
            model_name='customer',
            name='language',
            field=shark.utils.fields.LanguageField(blank=True, max_length=5, verbose_name='language'),
        ),
        migrations.AlterField(
            model_name='customer',
            name='payment_type',
            field=models.CharField(choices=[('invoice', 'Invoice'), ('direct_debit', 'Direct debit')], default='invoice', max_length=20, verbose_name='Payment Type'),
        ),
        migrations.AlterField(
            model_name='customer',
            name='tags',
            field=taggit.managers.TaggableManager(blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AlterField(
            model_name='customercontact',
            name='gender',
            field=models.CharField(blank=True, choices=[('M', 'Male'), ('F', 'Female')], max_length=1, verbose_name='Gender'),
        ),
        migrations.AlterField(
            model_name='customercontact',
            name='title',
            field=models.CharField(blank=True, choices=[('Herr', 'Herr'), ('Frau', 'Frau'), ('Fräulein', 'Fräulein'), ('Dr.', 'Dr.'), ('Dr. Dr.', 'Dr. Dr.'), ('Prof. Dr.', 'Prof. Dr.')], help_text='z.B. Herr, Frau, Dr., Prof.,...', max_length=20, verbose_name='Salutation'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import shark.utils.fields
import shark.utils.id_generators


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='invoice',
            name='payment_type',
            field=models.CharField(default=b'invoice', help_text="Will be copied from customer's preferences automatically.", max_length=20, verbose_name='Payment Type', choices=[(b'invoice', 'Invoice'), (b'direct_debit', 'Direct debit')]),
        ),
        migrations.AddField(
            model_name='invoiceitem',
            name='type',
            field=models.CharField(default=b'item', max_length=10, choices=[(b'item', b'Item'), (b'title', b'Title')]),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='language',
            field=shark.utils.fields.LanguageField(blank=True, help_text='This field will be automatically filled with the language of the customer. If no language for the customer is specified the default language (en-us) will be used.', max_length=5, verbose_name='language', choices=[(b'af', b'Afrikaans'), (b'ar', b'Arabic'), (b'ast', b'Asturian'), (b'az', b'Azerbaijani'), (b'bg', b'Bulgarian'), (b'be', b'Belarusian'), (b'bn', b'Bengali'), (b'br', b'Breton'), (b'bs', b'Bosnian'), (b'ca', b'Catalan'), (b'cs', b'Czech'), (b'cy', b'Welsh'), (b'da', b'Danish'), (b'de', b'German'), (b'el', b'Greek'), (b'en', b'English'), (b'en-au', b'Australian English'), (b'en-gb', b'British English'), (b'eo', b'Esperanto'), (b'es', b'Spanish'), (b'es-ar', b'Argentinian Spanish'), (b'es-mx', b'Mexican Spanish'), (b'es-ni', b'Nicaraguan Spanish'), (b'es-ve', b'Venezuelan Spanish'), (b'et', b'Estonian'), (b'eu', b'Basque'), (b'fa', b'Persian'), (b'fi', b'Finnish'), (b'fr', b'French'), (b'fy', b'Frisian'), (b'ga', b'Irish'), (b'gl', b'Galician'), (b'he', b'Hebrew'), (b'hi', b'Hindi'), (b'hr', b'Croatian'), (b'hu', b'Hungarian'), (b'ia', b'Interlingua'), (b'id', b'Indonesian'), (b'io', b'Ido'), (b'is', b'Icelandic'), (b'it', b'Italian'), (b'ja', b'Japanese'), (b'ka', b'Georgian'), (b'kk', b'Kazakh'), (b'km', b'Khmer'), (b'kn', b'Kannada'), (b'ko', b'Korean'), (b'lb', b'Luxembourgish'), (b'lt', b'Lithuanian'), (b'lv', b'Latvian'), (b'mk', b'Macedonian'), (b'ml', b'Malayalam'), (b'mn', b'Mongolian'), (b'mr', b'Marathi'), (b'my', b'Burmese'), (b'nb', b'Norwegian Bokmal'), (b'ne', b'Nepali'), (b'nl', b'Dutch'), (b'nn', b'Norwegian Nynorsk'), (b'os', b'Ossetic'), (b'pa', b'Punjabi'), (b'pl', b'Polish'), (b'pt', b'Portuguese'), (b'pt-br', b'Brazilian Portuguese'), (b'ro', b'Romanian'), (b'ru', b'Russian'), (b'sk', b'Slovak'), (b'sl', b'Slovenian'), (b'sq', b'Albanian'), (b'sr', b'Serbian'), (b'sr-latn', b'Serbian Latin'), (b'sv', b'Swedish'), (b'sw', b'Swahili'), (b'ta', b'Tamil'), (b'te', b'Telugu'), (b'th', b'Thai'), (b'tr', b'Turkish'), (b'tt', b'Tatar'), (b'udm', b'Udmurt'), (b'uk', b'Ukrainian'), (b'ur', b'Urdu'), (b'vi', b'Vietnamese'), (b'zh-cn', b'Simplified Chinese'), (b'zh-hans', b'Simplified Chinese'), (b'zh-hant', b'Traditional Chinese'), (b'zh-tw', b'Traditional Chinese')]),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='number',
            field=shark.utils.id_generators.IdField(unique=True, max_length=32, verbose_name='number', blank=True),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='type',
            field=models.CharField(default=b'invoice', max_length=20, verbose_name='type', choices=[(b'invoice', 'Invoice'), (b'correction', 'Correction of invoice')]),
        ),
    ]



# Generated by Django 2.2.2 on 2019-06-05 14:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('customer', '0005_customercontact_customer'),
    ]

    operations = [
        migrations.RenameField(
            model_name='customeraddress',
            old_name='address_addition',
            new_name='address_addition_1'
        ),
        migrations.AlterField(
            model_name='customeraddress',
            name='address_addition_1',
            field=models.CharField(default='', max_length=100, verbose_name='address addition (1st row)'),
        ),
        migrations.AddField(
            model_name='customeraddress',
            name='address_addition_2',
            field=models.CharField(default='', max_length=100, verbose_name='address addition (2nd row)'),
            preserve_default=False,
        ),
    ]



"""sharky URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/2.1/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.conf import settings
from django.contrib import admin
from django.urls import path, include

urlpatterns = [
    path('admin/billing/', include('shark.billing.admin_urls', namespace='billing_admin')),
    path('admin/sepa/', include('shark.sepa.admin_urls', namespace='sepa_admin')),
    path('admin/doc/', include('django.contrib.admindocs.urls')),
    path('admin/', admin.site.urls),
    path('admin_tools/', include('admin_tools.urls')),
    path('accounting/', include('shark.accounting.urls', namespace='accounting')),
    path('api/', include('shark.api_urls', namespace='api')),
    path('issue/', include('shark.issue.urls', namespace='issue')),
]

if settings.DEBUG:
    # Serve media files when DEBUG is enabled
    from django.conf.urls.static import static
    urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)



import argparse
from decimal import Decimal
from functools import partial

from django.apps import apps
from django.core.exceptions import ValidationError
from django.core.management.base import BaseCommand, CommandError
import unicodecsv as csv

from shark.customer.models import Customer
from shark.sepa.models import DirectDebitMandate


class Command(BaseCommand):
    args = '<csv_file>'
    help = 'import questions from CSV file'
    can_import_settings = True

    def add_arguments(self, parser):
        parser.add_argument('filename', type=argparse.FileType('r'))

    def handle(self, *args, **options):
        mandates = []
        fh = options['filename']
        reader = csv.DictReader(fh)
        has_error = False
        for row_index, row in enumerate(reader, 1):
            if row.get('#', '') == '#':
                # skip comment rows
                continue
            try:
                customer_number = row['customer_number']
                customer = Customer.objects.get(number=customer_number)
            except Customer.DoesNotExist:
                print('Unknown customer number: ', customer_number)
                continue
            mandate_data = {
                'customer': customer,
                'type': row['type'],
                'name': row['name'],
                'street': row['street'],
                'postal_code': row['postal_code'],
                'city': row['city'],
                'bank_name': row['bank_name'],
                'bic': row['bic'],
                'iban': row['iban'],
                'signed': row['signed'],
                'reference': customer.number + '-M01',
            }
            mandate = DirectDebitMandate(**mandate_data)
            try:
                mandate.clean_fields()
                mandate.clean()
            except ValidationError as e:
                has_error = True
                print('Errors in row %d:' % (row_index+1))
                for error in e:
                    print('  - %s: %s' % error)
                #print('Invalid row[%d]: %s' % (row_index, e))
            mandates.append(mandate_data)
        if has_error:
            print('Error(s) in document found. Aborting.')
            return
        for mandate_data in mandates:
            DirectDebitMandate.objects.get_or_create(**mandate_data)
            customer = mandate_data['customer']
            customer.payment_type = 'direct_debit'
            customer.save()



from django.db import models
from django.utils.translation import ugettext_lazy as _


class Account(models.Model):
    name = models.CharField(_('account name'),
            max_length=50)


class Transaction(models.Model):
    account = models.ForeignKey('banking.Account', on_delete=models.CASCADE)
    entry_date = models.DateField(_('entry date'))
    value_date = models.DateField(_('value date'))
    text_key = models.CharField(max_length=100) # XXX max_length?
    primanota = models.CharField(max_length=100) # XXX max_length?
    account_holder = models.CharField(max_length=100) # XXX max_length?
    account_number = models.CharField(max_length=100) # XXX max_length?
    bank_code = models.CharField(max_length=100) # XXX max_length?
    reference = models.TextField(max_length=100) # XXX max_length?
    currency = models.CharField(max_length=3, default='EUR')
    amount = models.DecimalField(max_digits=11, decimal_places=2)
    debit_credit = models.CharField(max_length=2)



from rest_framework import generics

from . import models
from . import api_serializers as serializers


class InvoiceList(generics.ListAPIView):
    queryset = models.Invoice.objects.all()
    serializer_class = serializers.InvoiceListSerializer


class InvoiceCreate(generics.CreateAPIView):
    queryset = models.Invoice.objects.all()
    serializer_class = serializers.InvoiceCreateSerializer


class InvoiceDetail(generics.RetrieveUpdateDestroyAPIView):
    queryset = models.Invoice.objects.all()
    serializer_class = serializers.InvoiceDetailSerializer



# Generated by Django 2.2.2 on 2019-06-05 14:10

from django.db import migrations


country_map = {
    'SCHWEIZ': 'ch',
    'UNITED KINGDOM': 'gb',
    'DEUTSCHLAND': 'de'
}


def convert_data(apps, schema_editor):
    Customer = apps.get_model('customer.Customer')

    for customer in Customer.objects.exclude(address=''):
        lines = [l.strip() for l in customer.address.split('\n')]
        # auto detect country
        if lines[-1].upper() in country_map:
            country = country_map[lines[-1].upper()]
            lines = lines[:-1]
        else:
            country = 'DE'
        try:
            postal_code, city = lines[-1].split(' ', 1)
        except ValueError:
            state = lines[-1]
            lines = lines[:-1]
        else:
            state = ''
        assert len(lines) <= 6
        customer.customeraddress_set.create(
            name=lines[0],
            address_addition_1=lines[1] if len(lines) >= 4 else '',
            address_addition_2=lines[2] if len(lines) >= 5 else '',
            street=lines[-2],
            postal_code=lines[-1].split(' ', 1)[0],
            city=lines[-1].split(' ', 1)[1],
            state=state,
            country=country,
            invoice_address=True,
        )
        customer.name = lines[0]
        customer.save()


class Migration(migrations.Migration):

    dependencies = [
        ('customer', '0007_add_state'),
    ]

    operations = [
        migrations.RunPython(convert_data)
    ]



from datetime import date

from django.test import TestCase

from shark.customer.models import Customer
from shark.billing.models import Invoice
from shark.utils.id_generators import DaysSinceEpoch
from shark.utils.id_generators import YearCustomerN
from shark.utils.id_generators import CustomerYearN


class DaysSinceEpochTestCase(TestCase):

    def test_next(self):
        gen = DaysSinceEpoch(Customer, 'number')
        customer = Customer.objects.create(number=gen.next())
        self.assertLess(customer.number, gen.next())

    def test_id_field(self):
        # first customer
        customer1 = Customer.objects.create()
        self.assertNotEqual(customer1.number, '')
        # second customer
        customer2 = Customer.objects.create()
        self.assertNotEqual(customer2.number, '')
        self.assertNotEqual(customer2.number, customer1.number)


class YearCustomerNTestCase(TestCase):

    def test_next(self):
        customer = Customer.objects.create(number='JOHNDOE')
        gen = YearCustomerN()
        gen.model_class = Invoice
        gen.field_name = 'number'
        today = date.today()
        prefix = '{:>04d}-JOHNDOE'.format(today.year)
        invoice1 = Invoice(customer=customer)
        invoice1.number = gen.next(invoice1)
        invoice1.save()
        self.assertEqual('%s-01' % prefix, invoice1.number)
        invoice2 = Invoice(customer=customer)
        invoice2.number = gen.next(invoice2)
        invoice2.save()
        self.assertEqual('%s-02' % prefix, invoice2.number)

    def test_invoice(self):
        customer = Customer.objects.create(number='JANEDOE')
        invoice = Invoice.objects.create(customer=customer)
        today = date.today()
        prefix = '{:>04d}-JANEDOE'.format(today.year)
        self.assertEqual('%s-01' % prefix, invoice.number)


class CustomerYearNTestCase(TestCase):

    def test_next(self):
        customer = Customer.objects.create(number='JOHNDOE')
        gen = CustomerYearN()
        gen.model_class = Invoice
        gen.field_name = 'number'
        today = date.today()
        prefix = 'JOHNDOE-%04d' % today.year
        invoice1 = Invoice(customer=customer)
        invoice1.number = gen.next(invoice1)
        invoice1.save()
        self.assertEqual('%s-01' % prefix, invoice1.number)
        invoice2 = Invoice(customer=customer)
        invoice2.number = gen.next(invoice2)
        invoice2.save()
        self.assertEqual('%s-02' % prefix, invoice2.number)

    def test_next_year2_nosep2(self):
        customer = Customer.objects.create(number='JOHNDOE')
        gen = CustomerYearN(year_length=2, separator2='')
        gen.model_class = Invoice
        gen.field_name = 'number'
        today = date.today()
        prefix = 'JOHNDOE-%s' % (str(today.year)[-2:])
        invoice1 = Invoice(customer=customer)
        invoice1.number = gen.next(invoice1)
        invoice1.save()
        self.assertEqual('%s01' % prefix, invoice1.number)
        invoice2 = Invoice(customer=customer)
        invoice2.number = gen.next(invoice2)
        invoice2.save()
        self.assertEqual('%s02' % prefix, invoice2.number)



import inspect

from django.utils import unittest


def import_testcases(module_name):
    '''
    Import test cases of a child module by adding it to this module
    prefixed with the name of the module.

    e.g. shark.foo.tests.FooTestCase becomes shark.tests.foo_FooTestCase
    and is then found by the test runner.
    '''
    shark_module = __import__('shark.%s.tests' % module_name)
    tests_module = getattr(shark_module, module_name).tests
    for name, value in tests_module.__dict__.items():
        if inspect.isclass(value) and issubclass(value, unittest.TestCase):
            globals()['%s_%s' % (module_name, name)] = value


import_testcases('utils')



# -*- coding: UTF-8 -*-

from datetime import date
from datetime import timedelta

from admin_tools.dashboard.modules import DashboardModule
from django.conf import settings
from django.db.models import Sum
from django.urls import reverse
from django.utils.translation import ugettext as _

from shark.billing.models import Invoice, InvoiceItem


class UnpaidInvoicesDashboardModule(DashboardModule):

    def __init__(self, **kwargs):
        super(UnpaidInvoicesDashboardModule, self).__init__(**kwargs)

        self.template = kwargs.get('template', 'billing/dashboard/unpaid-invoices.html')
        self.display = kwargs.get('display', 'tabs')
        self.layout = kwargs.get('layout', 'stacked')
        self.title = kwargs.get('title', _('Unpaid invoices'))

        self.is_empty = False

        today = date.today()
        two_weeks_ago = today - timedelta(days=14)
        thirty_days_ago = today - timedelta(days=30)

        admin_url = reverse('admin:%s_changelist' % settings.SHARK['MODELS']['billing.Invoice'].lower().replace('.', '_'))
        def get_admin_url(**kwargs):
            kwargs.setdefault('paid__isnull', True)
            return '%s?%s' % (admin_url, '&'.join('%s=%s' % item for item in kwargs.items()))

        unpaid = Invoice.objects.filter(paid__isnull=True)
        unpaid_lt14d = unpaid.filter(created__gt=two_weeks_ago)
        unpaid_gt14d = unpaid.filter(created__lte=two_weeks_ago).filter(created__gt=thirty_days_ago)
        unpaid_gt30d = unpaid.filter(created__lte=thirty_days_ago)

        self.unpaid = {
            'lt14d': {
                'count': unpaid_lt14d.count(),
                'url': get_admin_url(created__gt=two_weeks_ago),
            },
            'gt14d': {
                'count': unpaid_gt14d.count(),
                'url': get_admin_url(created__lte=two_weeks_ago, created__gt=thirty_days_ago),
            },
            'gt30d': {
                'count': unpaid_gt30d.count(),
                'url': get_admin_url(created__lte=thirty_days_ago),
            },
            'total': {
                'count': unpaid.count(),
                'url': get_admin_url(),
            }
        }


class LooseItemsDashboardModule(DashboardModule):

    def __init__(self, **kwargs):
        super(LooseItemsDashboardModule, self).__init__(**kwargs)

        self.template = kwargs.get('template', 'billing/dashboard/loose-items.html')
        self.display = kwargs.get('display', 'tabs')
        self.layout = kwargs.get('layout', 'stacked')
        self.title = kwargs.get('title', _('Loose invoice items'))

        self.is_empty = False

        self.items = InvoiceItem.objects.filter(invoice__isnull=True)
        self.item_count = self.items.count()
        self.list_url = reverse('admin:%s_changelist' % settings.SHARK['MODELS']['billing.InvoiceItem'].lower().replace('.', '_')) \
                + '?invoice__isnull=True'
        # XXX disabled for now
        #self.invoice_url = reverse('billing_admin:invoiceitem_invoice')



import tempfile

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import permission_required
from django.http import HttpResponse
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from django.utils.formats import date_format
from django.utils.translation import ugettext
from django.utils.translation import ungettext
from django.utils.translation import override as trans_override
from PyPDF2 import PdfFileWriter, PdfFileReader

from shark.billing.admin_forms import ImportItemsForm
from shark.billing.models import InvoiceItem
from shark.billing.models import Invoice

INVOICE_TERMS = settings.SHARK['INVOICE']['TERMS']


@permission_required('billing.add_invoice')
def invoice(request):
    items = InvoiceItem.objects.filter(invoice=None)
    return TemplateResponse(request, 'billing/admin/invoiceitem_invoice.html', {
        'items': items
    })


@permission_required('billing.change_invoice')
def invoice_pdf(request, number, correction=False):
    invoice = get_object_or_404(Invoice, number=number)
    if correction:
        invoice = invoice.correction
    from reportlab.lib.units import mm
    from reportlab.platypus import Paragraph
    from reportlab.platypus.flowables import Spacer
    from reportlab.platypus.flowables import KeepTogether

    from dinbrief.document import Document
    from dinbrief.invoice import ItemTable, TotalTable
    from dinbrief.styles import styles
    from dinbrief.template import BriefTemplate

    with trans_override(invoice.language):

        response = HttpResponse(content_type='application/pdf')
        if 'download' in request.GET:
            filename = '%s.pdf' % invoice.number
            response['Content-Disposition'] = 'attachment; filename=%s' % filename

        if invoice.type == Invoice.TYPE_INVOICE:
            if callable(INVOICE_TERMS):
                terms = INVOICE_TERMS(invoice)
            else:
                terms = [
                    Paragraph(term, styles['Terms'])
                    for term in INVOICE_TERMS
                ]
        else:
            terms = []

        template = BriefTemplate()
        document = Document(
            sender=invoice.sender_lines,
            recipient=invoice.recipient_lines,
            date=date_format(invoice.created, 'SHORT_DATE_FORMAT'),
            content=[
                Paragraph('%s %s' % (invoice.get_type_display() if not correction else ugettext(u'Correction of invoice'), invoice.number),
                        styles['Subject']),
                Spacer(template.CONTENT_WIDTH, 2*mm),
                ItemTable(template, invoice),
                KeepTogether(TotalTable(template, invoice)),
                Spacer(template.CONTENT_WIDTH, 10*mm),
            ] + terms)

        if settings.SHARK['INVOICE']['BACKGROUND']:
            with tempfile.TemporaryFile() as tmp:
                # Create content in a temporary file
                template.render(document, tmp)
                # Combine background with the content
                writer = PdfFileWriter()
                content = PdfFileReader(tmp)
                info_dict = writer._info.getObject()
                info_dict.update(content.getDocumentInfo())
                first_bg = PdfFileReader(open(
                        settings.SHARK['INVOICE']['BACKGROUND']['FIRST_PAGE'], 'rb'))
                later_bg = PdfFileReader(open(
                        settings.SHARK['INVOICE']['BACKGROUND']['LATER_PAGE'], 'rb'))
                bg = [first_bg.getPage(0), later_bg.getPage(0)]
                for i, page in enumerate(content.pages):
                    page.mergePage(bg[min(i, 1)])
                    page.compressContentStreams()
                    writer.addPage(page)
                writer.write(response)
        else:
            # Render content directly to the HTTP response object if no
            # background images are configured.
            template.render(document, response)

    return response


@permission_required('billing.change_invoice')
def correction_pdf(request, number):
    return invoice_pdf(request, number, correction=True)


@permission_required('billing.add_invoiceitem')
def import_items(request):
    if request.method == 'POST':
        form = ImportItemsForm(request.POST, request.FILES)
        if form.is_valid():
            items = form.cleaned_data['items']
            for item in items:
                item.save()
            messages.success(request, ungettext(
                '%(count)d invoice item imported.',
                '%(count)d invoice items imported.',
                len(items)
            ) % { 'count': len(items) })
    else:
        form = ImportItemsForm()
    from django.contrib.admin.helpers import AdminForm
    return TemplateResponse(request, 'billing/admin/import_items.html', {
        'form': form,
        'adminform': AdminForm(
            form=form,
            fieldsets=[
                (None, {
                    'fields': form.fields
                })
            ],
            prepopulated_fields={}),
        # The admin templates depend on those variables
        'add': True,
        'has_file_field': True,
        'errors': form.errors,
    })



"""
Django settings for shark project.

Generated by 'django-admin startproject' using Django 2.1.1.

For more information on this file, see
https://docs.djangoproject.com/en/2.1/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.1/ref/settings/
"""

import os
from django.utils.text import format_lazy
from django.utils.translation import ugettext_lazy as _

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'l_8$c18y1=urm5+(0)6lijec%&a9+%n+i4ufe6y)h&j&)bj$6e'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = (
    # Shark applications
    'shark',
    'shark.base',
    'shark.banking',
    'shark.billing',
    'shark.customer',
    'shark.documents',
    'shark.issue',
    'shark.project',
    'shark.sepa',
    'shark.vendor',
    # Admin tools
    'admin_tools',
    'admin_tools.theming',
    'admin_tools.menu',
    'admin_tools.dashboard',
    # Django contrib
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    # Misc
    'dinbrief',
    'taggit',
    'localflavor',
    'rest_framework',
)

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'shark.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
            'loaders': [
                'django.template.loaders.app_directories.Loader',
                'admin_tools.template_loaders.Loader',
            ],
        },
    },
]

WSGI_APPLICATION = 'shark.wsgi.application'


# Database
# https://docs.djangoproject.com/en/2.1/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}
'''
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.', # Add 'postgresql_psycopg2', 'mysql', 'sqlite3' or 'oracle'.
        'NAME': '',                      # Or path to database file if using sqlite3.
        'USER': '',                      # Not used with sqlite3.
        'PASSWORD': '',                  # Not used with sqlite3.
        'HOST': '',                      # Set to empty string for localhost. Not used with sqlite3.
        'PORT': '',                      # Set to empty string for default. Not used with sqlite3.
    }
}
'''


# Password validation
# https://docs.djangoproject.com/en/2.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/2.1/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'Europe/Berlin'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.1/howto/static-files/

STATIC_URL = '/static/'
STATIC_ROOT = os.path.join(BASE_DIR, 'htdocs', 'static')
MEDIA_URL = '/media/'
MEDIA_ROOT = os.path.join(BASE_DIR, 'htdocs', 'media')

FORMAT_MODULE_PATH = 'shark.base.formats'

ADMIN_TOOLS_INDEX_DASHBOARD = 'shark.dashboard.CustomIndexDashboard'
ADMIN_TOOLS_APP_INDEX_DASHBOARD = 'shark.dashboard.CustomAppIndexDashboard'

REST_FRAMEWORK = {
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.BasicAuthentication',
        'rest_framework.authentication.SessionAuthentication',
    ),
    'DEFAULT_PERMISSION_CLASSES': (
        'rest_framework.permissions.IsAdminUser',
    )
}

from decimal import Decimal
from datetime import timedelta

SHARK = {
    'VAT_RATE_CHOICES': (
        (Decimal('0.19'), '19%'),
        (Decimal('0.07'), '7%'),
        (Decimal('0.00'), format_lazy('0% ({})', _('tax free'))),
    ),
    'CUSTOMER': {
        'NUMBER_GENERATOR': 'shark.utils.id_generators.InitialAsNumber',
        'TYPE_CHOICES': [('default', _('Default'))],
        'TYPE_DEFAULT': 'default',
    },
    'INVOICE': {
        'BACKGROUND': {
            # 'FIRST_PAGE': ...
            # 'LATER_PAGE': ...
        },
        'SENDER': {
            "name": "settings.SHARK['INVOICE']['SENDER']['name']",
            "street": "settings.SHARK['INVOICE']['SENDER']['street']",
            "postal_code": "settings.SHARK['INVOICE']['SENDER']['postal_code']",
            "city": "settings.SHARK['INVOICE']['SENDER']['city']",
        },
        'TERMS': [
            "settings.SHARK['INVOICE']['TERMS']",
        ],
        'NUMBER_GENERATOR': 'shark.utils.id_generators.YearCustomerN',
        'UNIT_CHOICES': [
            ('s', _('second [s]')),
            ('min', _('minute [min]')),
            ('h', _('hour [h]')),
            ('d', _('day [d]')),
            ('w', _('week [w]')),
            ('m', _('month [m]')),
            ('a', _('year [a]')),
        ],
        'PAYMENT_TIMEFRAME': timedelta(days=14),
    },
    'SEPA': {
        'CREDITOR_ID': '',
        'CREDITOR_NAME': '',
        'CREDITOR_COUNTRY': 'DE',
        'CREDITOR_IBAN': '',
        'CREDITOR_BIC': '',
        'DEFAULT_MANDATE_TYPE': 'CORE',
        'TRANSACTION_REFERENCE_PREFIX': '',
        'PRE_NOTIFICATION_EMAIL_FROM': '',
        'PRE_NOTIFICATION_EMAIL_BCC': [],
    },
}



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sepa', '0004_optional_reference_and_executed'),
    ]

    operations = [
        migrations.AlterField(
            model_name='directdebitbatch',
            name='due_date',
            field=models.DateField(help_text='Must be min. 5 TARGET dates in the future for the first transaction and 2 target days in the future for recurring transactions.', verbose_name='due date'),
        ),
    ]



from django.db import models
from django.utils.translation import ugettext_lazy as _


class Project(models.Model):
    name = models.CharField(_('name'), max_length=100)
    customer = models.ForeignKey('customer.Customer', verbose_name=_('customer'), on_delete=models.CASCADE)
    active = models.BooleanField(_('active'), default=True)

    def __str__(self):
        return self.name



from rest_framework import serializers

from . import models


class CustomerSerializer(serializers.ModelSerializer):

    class Meta:
        model = models.Customer
        fields = ('number', 'address', 'language', 'created', 'updated')



from composite_field import CompositeField
from django.conf import settings
from django.db import models
from django.utils.functional import curry
from django.utils.translation import ugettext_lazy as _
from django_countries.fields import CountryField


class OldAddressField(models.TextField):

    def contribute_to_class(self, cls, name):
        super().contribute_to_class(cls, name)
        getter = curry(self._get_FIELD_lines)
        setter = curry(self._set_FIELD_lines)
        cls.add_to_class('get_%s_lines' % name, getter)
        cls.add_to_class('set_%s_lines' % name, setter)
        cls.add_to_class('%s_lines' % name, property(getter, setter))

    def _get_FIELD_lines(self, obj):
        return self.value_from_object(obj).split('\n')

    def _set_FIELD_lines(self, obj, value):
        setattr(obj, self.attname, '\n'.join(value))


class AddressField(CompositeField):
    name = models.CharField(_('name'), max_length=100)
    address_addition_1 = models.CharField(_('address addition (1st row)'), max_length=100, blank=True)
    address_addition_2 = models.CharField(_('address addition (2nd row)'), max_length=100, blank=True)
    street = models.CharField(_('street'), max_length=100)
    city = models.CharField(_('city'), max_length=100)
    postal_code = models.CharField(_('postal code'), max_length=10)
    state = models.CharField(max_length=100, blank=True)
    country = CountryField(_('country'))

    def __init__(self, **kwargs):
        blank = kwargs.pop('blank', False)
        super().__init__(**kwargs)
        if blank:
            self['name'].blank = blank
            self['street'].blank = blank
            self['city'].blank = blank
            self['postal_code'].blank = blank
            self['country'].blank = blank


class LanguageField(models.CharField):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_length', 5)
        kwargs.setdefault('choices', settings.LANGUAGES)
        kwargs.setdefault('help_text', _('This field will be automatically filled with the language of the customer. If no language for the customer is specified the default language (%s) will be used.' % settings.LANGUAGE_CODE))
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["choices"]
        del kwargs['help_text']
        if 'default' in kwargs:
            del kwargs["default"]
        return name, path, args, kwargs



# Generated by Django 2.2.2 on 2019-06-05 12:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('issue', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='attachment',
            name='file',
            field=models.FileField(upload_to='issue/attachment', verbose_name='file'),
        ),
        migrations.AlterField(
            model_name='relation',
            name='type',
            field=models.CharField(choices=[('parent', 'parent'), ('related', 'related'), ('duplicate', 'duplicate')], max_length=10, verbose_name='type'),
        ),
    ]



from django.contrib import admin

from . import models


@admin.register(models.Project)
class ProjectAdmin(admin.ModelAdmin):
    list_display = ('name', 'customer', 'active')
    list_editable = ('active',)
    list_filter = ('customer', 'active')
    search_fields = ('name',)
    autocomplete_fields = ('customer',)



from django.urls import reverse
from rest_framework import serializers

from shark.customer.api_serializers import CustomerSerializer
from shark.customer.models import Customer

from . import models

class InvoiceItemSerializer(serializers.ModelSerializer):

    class Meta:
        model = models.InvoiceItem
        fields = ('id', 'position', 'quantity', 'sku', 'text',
                'begin', 'end', 'price', 'unit', 'discount',
                'vat_rate')


class InvoiceDetailSerializer(serializers.ModelSerializer):
    #customer = serializers.HyperlinkedRelatedField( # XXX
    items = InvoiceItemSerializer(source='item_set', many=True)

    class Meta:
        model = models.Invoice
        fields = ('id', 'customer', 'type', 'number', 'language',
                'sender', 'recipient', 'net', 'gross', 'created',
                'reminded', 'paid', 'items')
        depth = 1


class CustomerField(serializers.Field):

    def to_representation(self, obj):
        return obj.number

    def to_internal_value(self, data):
        try:
            return Customer.objects.get(number=data)
        except Customer.DoesNotExist:
            raise RuntimeError('No such customer')


class InvoiceListSerializer(serializers.ModelSerializer):
    url = serializers.SerializerMethodField()
    customer = CustomerField()

    class Meta:
        model = models.Invoice
        fields = ('id', 'url', 'customer', 'type', 'number',
                'net', 'gross', 'created', 'reminded', 'paid')

    def get_url(self, instance):
        request = self.context['request']
        relative_url = reverse('api:billing:invoice_detail', kwargs={'pk': instance.pk})
        return request.build_absolute_uri(relative_url)


class InvoiceCreateSerializer(serializers.ModelSerializer):
    customer = CustomerSerializer()
    items = InvoiceItemSerializer(many=True)

    class Meta:
        model = models.Invoice
        fields = ('id', 'customer', 'type', 'number', 'language',
                'sender', 'recipient', 'net', 'gross', 'created',
                'reminded', 'paid', 'items')
        #depth = 1

    def create(self, validated_data):
        # update or create customer
        customer_data = validated_data.pop('customer')
        customer.objects.update_or_create(
            number=customer_data.pop('number'),
            defaults=customer_data
        )
        # create invoice
        invoice = models.Invoice.objects.create(**validated_data)
        # create invoice items
        items_data = validated_data.pop('items')
        for item_data in items_data:
            item = models.InvoiceItem(**item_data)
            invoice.item_set.add(item)
        # recalculate invoice totals
        invoice.recalculate()
        invoice.save()
        return invoice



from django.contrib import admin

from shark.banking import models


class AccountAdmin(admin.ModelAdmin):
    list_display = ('name',)


class TransactionAdmin(admin.ModelAdmin):
    list_display = ('account', 'entry_date', 'value_date', 'reference', 'amount', 'debit_credit')
    list_filter = ('account', 'entry_date', 'value_date')
    search_fields = ('reference',)


admin.site.register(models.Account, AccountAdmin)
admin.site.register(models.Transaction, TransactionAdmin)



"""
This file was generated with the customdashboard management command, it
contains the two classes for the main dashboard and app index dashboard.
You can customize these classes as you want.

To activate your index dashboard add the following to your settings.py::
    ADMIN_TOOLS_INDEX_DASHBOARD = 'shark.dashboard.CustomIndexDashboard'

And to activate the app index dashboard::
    ADMIN_TOOLS_APP_INDEX_DASHBOARD = 'shark.dashboard.CustomAppIndexDashboard'
"""

from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from admin_tools.dashboard import modules, Dashboard, AppIndexDashboard
from admin_tools.utils import get_admin_site_name

from shark.billing.dashboard import UnpaidInvoicesDashboardModule
from shark.billing.dashboard import LooseItemsDashboardModule


class CustomIndexDashboard(Dashboard):
    """
    Custom index dashboard for shark.
    """
    def init_with_context(self, context):
        site_name = get_admin_site_name(context)
        # append a link list module for "quick links"
        self.children.append(modules.LinkList(
            _('Quick links'),
            layout='inline',
            draggable=False,
            deletable=False,
            collapsible=False,
            children=[
                [_('Return to site'), '/'],
                [_('Change password'),
                 reverse('%s:password_change' % site_name)],
                [_('Log out'), reverse('%s:logout' % site_name)],
            ]
        ))

        self.children.append(UnpaidInvoicesDashboardModule())
        self.children.append(LooseItemsDashboardModule())

        # append an app list module for "Applications"
        self.children.append(modules.AppList(
            _('Applications'),
            exclude=('django.contrib.*',),
        ))

        # append an app list module for "Administration"
        self.children.append(modules.AppList(
            _('Administration'),
            models=('django.contrib.*',),
        ))

        # append a recent actions module
        self.children.append(modules.RecentActions(_('Recent Actions'), 5))


class CustomAppIndexDashboard(AppIndexDashboard):
    """
    Custom app index dashboard for shark.
    """

    # we disable title because its redundant with the model list module
    title = ''

    def __init__(self, *args, **kwargs):
        AppIndexDashboard.__init__(self, *args, **kwargs)

        # append a model list module and a recent actions module
        self.children += [
            modules.ModelList(self.app_title, self.models),
            modules.RecentActions(
                _('Recent Actions'),
                include_list=self.get_app_content_types(),
                limit=5
            )
        ]

    def init_with_context(self, context):
        """
        Use this method if you need to access the request context.
        """
        return super(CustomAppIndexDashboard, self).init_with_context(context)



'''
This module contains classes for generating special sequences like
customer numbers that are not plain integer fields.
'''

from copy import copy
from datetime import date

from django.db import models
from django.db.models import signals
from django.utils.functional import curry

from shark.utils.int2base import int2base


class IdGenerator(object):
    # This value should be overwritten by subclasses.
    # Typically an invoice number should not exceed 16 characters as a lot
    # of accounting software assumes this limit. 32 chars also allows the
    # use of UUIDs for an ID field which I assume is a sane maximum length.
    max_length = 32
    '''
    This is just an interface class and does not contain any implementation.
    '''

    def __init__(self, model_class=None, field_name=None):
        self.model_class = model_class
        self.model_class_given = model_class is not None
        self.field_name = field_name
        self.field_name_given = field_name is not None

    def get_queryset(self):
        return self.model_class.objects.all() \
                .order_by('-%s' % self.field_name)

    def get_last(self):
        obj = self.get_queryset()[:1].get()
        return getattr(obj, self.field_name)


class InitialAsNumber(IdGenerator):
    '''
    Generate numbers of the format <prefix><initial><n>

    This is typically used in German accounting for customer numbers.
    e.g. "Epic Greatness Corp" might get the customer number "10503".
    "1" being the prefix. "05" for the fifth letter in the alphabet and "03"
    as this company being the third customer with that first lettter in the
    name.

    prefix: defaults to ''
    initial: the first letter of the given initial field as number between
        01 and 26. 00 is used for non-letter initials.
    n: simple counter with n_length characters and to base n_base
    '''
    def __init__(self, model_class=None, field_name=None, prefix='',
            initial_field_name='name',
            n_length=2, n_base=10):
        super(InitialAsNumber, self).__init__(model_class, field_name)
        self.initial_length = 2
        self.initial_field_name = initial_field_name
        self.n_length = n_length
        self.n_base = n_base
        self.format_string = '{prefix}{initial:0>2s}{n:0>%ds}' % (n_length)
        self.max_length = len(prefix) + 2 + n_length

    def format(self, initial, n):
        return self.format_string.format(
            prefix=self.prefix,
            initial=self.format_initial(initial),
            n=self.format_n(n))

    def format_initial(self, s):
        initial = ord(s[0].lower() - ord('a') + 1)
        return f'{initial:0>2s}' if 1 <= initial <= 26 else '00'

    def parse_initial(self, s):
        return chr(ord('a') + int(s) - 1)

    def format_n(self, n):
        return int2base(n, self.n_base)

    def parse(self, s):
        lp = len(self.prefix)
        prefix = s[:lp]
        initial = self.parse_initial(s[lp:lp+2])
        n = s[lp+2:]
        return (prefix, initial, n)

    def get_start(self, initial):
        return self.format(initial, 1)

    def get_last(self, initial):
        start = self.prefix + self.format_initial(initial)
        obj = self.get_queryset() \
                .filter(**{f'{self.field_name}__istartswith': start}) \
                [:1].get()
        return getattr(obj, self.field_name)

    def next(self, instance=None):
        initial = getattr(instance, self.initial_field_name)[0]
        start = self.get_start(initial)
        try:
            last = self.get_last(initial)
            if start > last:
                return start
            (prefix, last_initial, last_n) = self.parse(last)
            return self.format(days, last_n+1)
        except self.model_class.DoesNotExist:
            return start


class DaysSinceEpoch(IdGenerator):
    '''
    Generate numbers of the format <prefix><days><n>

    prefix: defaults to ''
    days: days since epoch (1970-01-01)
    n: simple counter with n_length characters and to base n_base
    '''

    def __init__(self, model_class=None, field_name=None, prefix='',
            epoch=date(1970, 1, 1), days_length=5, days_base=10,
            n_length=3, n_base=10):
        super(DaysSinceEpoch, self).__init__(model_class, field_name)
        self.days_length = days_length
        self.days_base = days_base
        self.prefix = prefix
        self.n_length = n_length
        self.n_base = n_base
        self.epoch = epoch
        self.format_string = '{prefix}{days:0>%ds}{n:0>%ds}' % (
                days_length, n_length)
        self.max_length = len(prefix) + days_length + n_length

    def format(self, days, n):
        return self.format_string.format(
            prefix=self.prefix,
            days=self.format_days(days),
            n=self.format_n(n))

    def format_days(self, days):
        return int2base(days, self.days_base)

    def format_n(self, n):
        return int2base(n, self.n_base)

    def parse(self, s):
        lp = len(self.prefix)
        ld = self.days_length
        return (s[:lp], int(s[lp:lp+ld], self.days_base),
                int(s[lp+ld:], self.n_base))

    def get_start(self, today=None):
        today = today or date.today()
        days = (today - self.epoch).days
        return self.format(days, 0)

    def next(self, instance=None, today=None):
        start = self.get_start(today)
        try:
            last = self.get_last()
            if start > last:
                return start
            (prefix, days, last_n) = self.parse(last)
            return self.format(days, last_n+1)
        except self.model_class.DoesNotExist:
            return start


class YearCustomerN(IdGenerator):
    '''
    Generate numbers of the format
    <prefix><year><separator1><customer_number><separator2><n>
    e.g. 2012-EXAMPLE-01

    prefix: defaults to ''
    separator1: defaults to '-'
    separator2: defaults to '-'
    n: simple counter with n_length characters to the base n_base
    '''

    def __init__(self, model_class=None, field_name=None, prefix='',
            separator1='-', separator2='-', customer_number_length=20, n_length=2, n_base=10):
        # FIXME Is there a way to figure out the customer number length
        #       automatically?
        super(YearCustomerN, self).__init__(model_class, field_name)
        self.prefix = prefix
        self.separator1 = separator1
        self.separator2 = separator2
        self.n_length = n_length
        self.n_base = n_base
        # <prefix><year><separator1><customer_number><separator2><n>
        self.year_customer_format_string = \
                '{prefix}{year:04d}{separator1}' \
                '{customer_number}{separator2}'
        self.format_string = self.year_customer_format_string + \
                '{n:0>%ds}' % n_length
        self.max_length = len(prefix) + len(separator1) + 4 + \
                customer_number_length + len(separator2) + n_length

    def format(self, year, customer_number, n):
        return self.format_string.format(
            prefix=self.prefix,
            year=year,
            separator1=self.separator1,
            separator2=self.separator2,
            customer_number=customer_number,
            n=self.format_n(n))

    def format_n(self, n):
        return int2base(n, self.n_base)

    def parse(self, s):
        lp = len(self.prefix)
        prefix, rest = (s[:lp], s[lp:])
        year, rest = rest.split(self.separator1, 1)
        year = int(year)
        customer_number, n = rest.rsplit(self.separator2, 1)
        n = int(n, self.n_base)
        return (year, customer_number, n)

    def get_start(self, customer, today=None):
        today = today or date.today()
        return self.format(today.year, customer.number, 1)

    def next(self, instance, today=None):
        customer = instance.customer
        today = today or date.today()
        start = self.get_start(customer, today)
        try:
            last = self.get_last(customer, today)
            if start > last:
                return start
            (year, customer_number, last_n) = self.parse(last)
            return self.format(year, customer_number, last_n+1)
        except self.model_class.DoesNotExist:
            return start

    def get_queryset(self, customer, today):
        prefix = self.year_customer_format_string.format(
            prefix=self.prefix,
            year=today.year,
            separator1=self.separator1,
            separator2=self.separator2,
            customer_number=customer.number)
        return self.model_class.objects.all() \
                .filter(**{ ('%s__startswith' % self.field_name): prefix }) \
                .order_by('-%s' % self.field_name)

    def get_last(self, customer, today):
        obj = self.get_queryset(customer, today)[:1].get()
        return getattr(obj, self.field_name)


class CustomerYearN(IdGenerator):
    '''
    Generate numbers of the format
    <prefix><customer_number><separator1><year><separator2><n>
    e.g. EXAMPLE-2012-01

    prefix: defaults to ''
    separator1: defaults to '-'
    separator2: defaults to '-', optional
    n: simple counter with n_length characters to the base n_base
    '''

    def __init__(self, model_class=None, field_name=None, prefix='',
            customer_number_length=20, separator1='-', year_length=4,
            separator2='-', n_length=2, n_base=10):
        super(CustomerYearN, self).__init__(model_class, field_name)
        self.prefix = prefix
        self.separator1 = separator1
        self.year_length = year_length
        self.separator2 = separator2
        self.n_length = n_length
        self.n_base = n_base
        # <prefix><year><separator1><customer_number><separator2><n>
        self.customer_year_format_string = \
                '{prefix}{customer_number}{separator1}' \
                '{year:0>%ds}{separator2}' % year_length
        self.format_string = self.customer_year_format_string + \
                '{n:0>%ds}' % n_length
        self.max_length = len(prefix) + len(separator1) + 4 + \
                customer_number_length + len(separator2) + n_length

    def format(self, customer_number, year, n):
        return self.format_string.format(
            prefix=self.prefix,
            year=self.format_year(year),
            separator1=self.separator1,
            separator2=self.separator2,
            customer_number=customer_number,
            n=self.format_n(n))

    def format_year(self, year):
        return (('%%0%dd' % self.year_length) % year)[-self.year_length:]

    def format_n(self, n):
        return int2base(n, self.n_base)

    def parse(self, s):
        lp = len(self.prefix)
        prefix, rest = (s[:lp], s[lp:])
        customer_number, rest = rest.split(self.separator1, 1)
        if self.separator2:
            year, n = rest.rsplit(self.separator2, 1)
        else:
            ly = self.year_length
            year, n = (rest[:ly], rest[ly:])
        year = int(year)
        n = int(n, self.n_base)
        return (customer_number, year, n)

    def get_start(self, customer, today=None):
        today = today or date.today()
        return self.format(customer.number, today.year, 1)

    def next(self, instance, today=None):
        customer = instance.customer
        today = today or date.today()
        start = self.get_start(customer, today)
        try:
            last = self.get_last(customer, today)
            if start > last:
                return start
            (customer_number, year, last_n) = self.parse(last)
            return self.format(customer_number, year, last_n+1)
        except self.model_class.DoesNotExist:
            return start

    def get_queryset(self, customer, today):
        prefix = self.customer_year_format_string.format(
            prefix=self.prefix,
            customer_number=customer.number,
            separator1=self.separator1,
            year=self.format_year(today.year),
            separator2=self.separator2)
        return self.model_class.objects.all() \
                .filter(**{ ('%s__startswith' % self.field_name): prefix }) \
                .order_by('-%s' % self.field_name)

    def get_last(self, customer, today):
        obj = self.get_queryset(customer, today)[:1].get()
        return getattr(obj, self.field_name)


class IdField(models.CharField):

    def __init__(self, **kwargs):
        self.generator = kwargs.pop('generator', None)
        kwargs.setdefault('max_length', 32)
        if self.generator:
            if self.generator.max_length > kwargs['max_length']:
                raise RuntimeError('The generator is capable of generating IDs exceeding the max_length of this field. Consider using a different generator class or setting a higher max_length value to this field.')
        kwargs.setdefault('blank', True)
        kwargs.setdefault('unique', True)
        super(IdField, self).__init__(**kwargs)

    def contribute_to_class(self, cls, name):
        super(IdField, self).contribute_to_class(cls, name)
        if self.generator:
            generator = copy(self.generator)
            if not generator.model_class_given:
                generator.model_class = cls
            if not generator.field_name_given:
                generator.field_name = name
            signals.pre_save.connect(curry(self._pre_save, generator=generator),
                    sender=cls, weak=False)

    # Do not name this method 'pre_save' as it will otherwise be called without
    # the generator argument.
    def _pre_save(self, generator, sender, instance, *args, **kwargs):
        if getattr(instance, self.name, ''):
            # Do not create an ID for objects that already have a value set.
            return
        value = generator.next(instance=instance)
        setattr(instance, self.name, value)



from xml.sax.saxutils import escape

from django.contrib import admin
from django.utils.html import format_html_join, mark_safe

from shark.customer import models


class CustomerAddressInline(admin.StackedInline):
    model = models.CustomerAddress
    extra = 0


class CustomerContactInline(admin.StackedInline):
    model = models.CustomerContact
    extra = 0


class CustomerCommentInline(admin.StackedInline):
    model = models.CustomerComment
    extra = 0


class CustomerAdmin(admin.ModelAdmin):
    list_display = ['number', 'address_html', 'created']
    list_filter = ['created']
    search_fields = ['number', 'address']
    date_hierarchy = 'created'
    inlines = [CustomerAddressInline, CustomerContactInline, CustomerCommentInline]

    def address_html(self, instance):
        return format_html_join(mark_safe('<br>'), '{}', ((line,) for line in instance.address_lines))
    address_html.short_description = models.Customer._meta.get_field('address').verbose_name
    address_html.admin_order_field = 'address'


if not models.Customer._meta.abstract:
    admin.site.register(models.Customer, CustomerAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import taggit.managers
import shark.utils.fields
import django_countries.fields


class Migration(migrations.Migration):

    dependencies = [
        ('taggit', '0001_initial'),
        ('customer', '0002_add_hourly_rate'),
    ]

    operations = [
        migrations.CreateModel(
            name='CustomerAddress',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=100, verbose_name='name')),
                ('address_addition', models.CharField(max_length=100, verbose_name='name')),
                ('street', models.CharField(max_length=100, verbose_name='street')),
                ('city', models.CharField(max_length=100, verbose_name='city')),
                ('postal_code', models.CharField(max_length=10, verbose_name='postal code')),
                ('country', django_countries.fields.CountryField(max_length=2, verbose_name='country')),
                ('invoice_address', models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name='CustomerComment',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('text', models.TextField()),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name='CustomerContact',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('gender', models.CharField(blank=True, max_length=1, verbose_name='Gender', choices=[(b'M', 'Male'), (b'F', 'Female')])),
                ('title', models.CharField(blank=True, help_text='z.B. Herr, Frau, Dr., Prof.,...', max_length=20, verbose_name='Salutation', choices=[(b'Herr', b'Herr'), (b'Frau', b'Frau'), (b'Fr\xc3\xa4ulein', b'Fr\xc3\xa4ulein'), (b'Dr.', b'Dr.'), (b'Dr. Dr.', b'Dr. Dr.'), (b'Prof. Dr.', b'Prof. Dr.')])),
                ('first_name', models.CharField(max_length=20, verbose_name='First name', blank=True)),
                ('last_name', models.CharField(max_length=20, verbose_name='Last name', blank=True)),
                ('phone_number', models.CharField(max_length=50, blank=True)),
                ('mobile_number', models.CharField(max_length=50, blank=True)),
                ('fax_number', models.CharField(max_length=50, blank=True)),
            ],
        ),
        migrations.AddField(
            model_name='customer',
            name='daily_rate',
            field=models.DecimalField(null=True, verbose_name='daily rate', max_digits=7, decimal_places=2, blank=True),
        ),
        migrations.AddField(
            model_name='customer',
            name='enabled',
            field=models.BooleanField(default=True),
        ),
        migrations.AddField(
            model_name='customer',
            name='invoice_dispatch_type',
            field=models.CharField(default=b'email', max_length=20, verbose_name='Invoice dispatch type', choices=[(b'email', 'via email'), (b'fax', 'via FAX'), (b'mail', 'via mail')]),
        ),
        migrations.AddField(
            model_name='customer',
            name='payment_type',
            field=models.CharField(default=b'invoice', max_length=20, verbose_name='Payment Type', choices=[(b'invoice', 'Invoice'), (b'direct_debit', 'Direct debit')]),
        ),
        migrations.AddField(
            model_name='customer',
            name='tags',
            field=taggit.managers.TaggableManager(to='taggit.Tag', through='taggit.TaggedItem', help_text='A comma-separated list of tags.', verbose_name='Tags'),
        ),
        migrations.AddField(
            model_name='customer',
            name='vatin',
            field=models.CharField(help_text='Value added tax identification number', max_length=14, verbose_name='VATIN', blank=True),
        ),
        migrations.AlterField(
            model_name='customer',
            name='language',
            field=shark.utils.fields.LanguageField(blank=True, max_length=5, verbose_name='language', choices=[(b'af', b'Afrikaans'), (b'ar', b'Arabic'), (b'ast', b'Asturian'), (b'az', b'Azerbaijani'), (b'bg', b'Bulgarian'), (b'be', b'Belarusian'), (b'bn', b'Bengali'), (b'br', b'Breton'), (b'bs', b'Bosnian'), (b'ca', b'Catalan'), (b'cs', b'Czech'), (b'cy', b'Welsh'), (b'da', b'Danish'), (b'de', b'German'), (b'el', b'Greek'), (b'en', b'English'), (b'en-au', b'Australian English'), (b'en-gb', b'British English'), (b'eo', b'Esperanto'), (b'es', b'Spanish'), (b'es-ar', b'Argentinian Spanish'), (b'es-mx', b'Mexican Spanish'), (b'es-ni', b'Nicaraguan Spanish'), (b'es-ve', b'Venezuelan Spanish'), (b'et', b'Estonian'), (b'eu', b'Basque'), (b'fa', b'Persian'), (b'fi', b'Finnish'), (b'fr', b'French'), (b'fy', b'Frisian'), (b'ga', b'Irish'), (b'gl', b'Galician'), (b'he', b'Hebrew'), (b'hi', b'Hindi'), (b'hr', b'Croatian'), (b'hu', b'Hungarian'), (b'ia', b'Interlingua'), (b'id', b'Indonesian'), (b'io', b'Ido'), (b'is', b'Icelandic'), (b'it', b'Italian'), (b'ja', b'Japanese'), (b'ka', b'Georgian'), (b'kk', b'Kazakh'), (b'km', b'Khmer'), (b'kn', b'Kannada'), (b'ko', b'Korean'), (b'lb', b'Luxembourgish'), (b'lt', b'Lithuanian'), (b'lv', b'Latvian'), (b'mk', b'Macedonian'), (b'ml', b'Malayalam'), (b'mn', b'Mongolian'), (b'mr', b'Marathi'), (b'my', b'Burmese'), (b'nb', b'Norwegian Bokmal'), (b'ne', b'Nepali'), (b'nl', b'Dutch'), (b'nn', b'Norwegian Nynorsk'), (b'os', b'Ossetic'), (b'pa', b'Punjabi'), (b'pl', b'Polish'), (b'pt', b'Portuguese'), (b'pt-br', b'Brazilian Portuguese'), (b'ro', b'Romanian'), (b'ru', b'Russian'), (b'sk', b'Slovak'), (b'sl', b'Slovenian'), (b'sq', b'Albanian'), (b'sr', b'Serbian'), (b'sr-latn', b'Serbian Latin'), (b'sv', b'Swedish'), (b'sw', b'Swahili'), (b'ta', b'Tamil'), (b'te', b'Telugu'), (b'th', b'Thai'), (b'tr', b'Turkish'), (b'tt', b'Tatar'), (b'udm', b'Udmurt'), (b'uk', b'Ukrainian'), (b'ur', b'Urdu'), (b'vi', b'Vietnamese'), (b'zh-cn', b'Simplified Chinese'), (b'zh-hans', b'Simplified Chinese'), (b'zh-hant', b'Traditional Chinese'), (b'zh-tw', b'Traditional Chinese')]),
        ),
        migrations.AddField(
            model_name='customercomment',
            name='customer',
            field=models.ForeignKey(to='customer.Customer', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='customeraddress',
            name='customer',
            field=models.ForeignKey(to='customer.Customer', on_delete=models.CASCADE),
        ),
    ]
